#!/usr/bin/env python2

# This file is part of Shoebot.
# Copyright (C) 2007-2009 the Shoebot authors
# See the COPYING file for the full license text.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#   The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
#   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
#   EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
Drawbot and Nodebot are similar grammars, so they both inherit from Bot
"""

import sys, os

from shoebot.data import BezierPath, EndClip, Color, Text, Variable, \
                         Image, ClippingPath, \
                         NUMBER, TEXT, BOOLEAN, BUTTON, \
                         ShoebotError, \
                         RGB, HSB, \
                         CENTER, CORNER, CORNERS

from grammar import Grammar

from pkg_resources import resource_filename, Requirement
from glob import glob
import random as r

import locale, gettext

SBOT_ROOT=resource_filename(Requirement.parse("shoebot"), "")
APP = 'shoebot'
DIR = sys.prefix + '/share/shoebot/locale'
locale.setlocale(locale.LC_ALL, '')
gettext.bindtextdomain(APP, DIR)
#gettext.bindtextdomain(APP)
gettext.textdomain(APP)
_ = gettext.gettext

import sys
LIB_DIRS = [
    os.path.join(SBOT_ROOT, 'local', 'share', 'shoebot', 'lib'),
    os.path.join(SBOT_ROOT, 'lib'),
    os.path.join(SBOT_ROOT, 'share', 'shoebot', 'lib')]
for LIB_DIR in LIB_DIRS:
    if os.path.isdir(LIB_DIR):
        sys.path.append(LIB_DIR)

TOP_LEFT = 1
BOTTOM_LEFT = 2


class Bot(Grammar):
    '''
    The Parts of the Grammar common to DrawBot, NodeBot and ShoeBot.
    '''
    RGB = RGB
    HSB = HSB    

    LEFT = 'left'
    RIGHT = 'right'

    CENTER = CENTER
    CORNER = CORNER
    CORNERS = CORNERS

    LEFT = 'left'
    RIGHT = 'right'
    JUSTIFY = 'justify'

    NUMBER = NUMBER
    TEXT = TEXT
    BOOLEAN = BOOLEAN
    BUTTON = BUTTON

    inch = 72
    cm = 28.3465
    mm = 2.8346

    # Default mouse values
    MOUSEX = -1
    MOUSEY = -1
    mousedown = False

    # Default key values
    key = '-'
    keycode = 0
    keydown = False

    def __init__(self, canvas, namespace=None, vars=None):
        '''
        :param canvas: Canvas implementation for output.
        :param namespace: Optionally specify a dict to inject as namespace
        :param vars: Optional dict containing initial values for variables
        '''
        Grammar.__init__(self, canvas, namespace=namespace, vars=vars)
        canvas.set_bot(self)

        self._autoclosepath = True
        self._path = None

        if self._input_device:
            # Get constants like KEY_DOWN, KEY_LEFT
            for key_name, value in self._input_device.get_key_map().items():
                self._namespace[key_name] = value
                setattr(self, key_name, value)

        self._canvas.size = None
        self._frame = 1
        self._set_initial_defaults() ### TODO Look at these
        
    def _set_initial_defaults(self):
        '''Set the default values. Called at __init__ and at the end of run(),
        do that new draw loop iterations don't take up values left over by the
        previous one.'''
        DEFAULT_WIDTH, DEFAULT_HEIGHT = self._canvas.DEFAULT_SIZE
        self.WIDTH = self._namespace.get('WIDTH', DEFAULT_WIDTH)
        self.HEIGHT = self._namespace.get('HEIGHT', DEFAULT_WIDTH)
        if 'WIDTH' in self._namespace or 'HEIGHT' in self._namespace:
            self.size(w=self._namespace.get('WIDTH'), h=self._namespace.get('HEIGHT'))

        self._transformmode = Bot.CENTER

        self._canvas.settings(
            fontfile = "assets/notcouriersans.ttf",
            fontsize = 16,
            align = Bot.LEFT,
            lineheight = 1,
            fillcolor = self.color(.2),
            strokecolor = None,
            strokewidth = 1.0,
            background = self.color(1, 1, 1))

    def _set_dynamic_vars(self):
        self._namespace['FRAME'] = self._frame


    # Get input

    def _mouse_button_down(self, button):
        '''GUI callback for mouse button down'''
        self._namespace['mousedown'] = True

    def _mouse_button_up(self, button):
        '''GUI callback for mouse button up'''
        self._namespace['mousedown'] = self._input_device.mouse_down

    def _mouse_pointer_moved(self, x, y):
        '''GUI callback for mouse moved'''
        self._namespace['MOUSEX'] = x
        self._namespace['MOUSEY'] = y

    def _key_pressed(self, key, keycode):
        '''GUI callback for key pressed'''
        self._namespace['key'] = key
        self._namespace['keycode'] = keycode
        self._namespace['keydown'] = True

    def _key_released(self, key, keycode):
        '''GUI callback for key released'''
        self._namespace['keydown'] = self._input_device.key_down

    #### Functions for override

    def setup(self):
        """ For override by user sketch """
        pass

    def draw(self):
        """ For override by user sketch """
        self._dynamic = False

    #### Classes

    def _makeInstance(self, clazz, args, kwargs):
        '''Creates an instance of a class defined in this document.
           This method sets the context of the object to the current context.'''
        inst = clazz(self, *args, **kwargs)
        return inst

    def _makeColorableInstance(self, clazz, args, kwargs):
        """
        Create an object, if fill, stroke or strokewidth
        is not specified, get them from the _canvas

        :param clazz:
        :param args:
        :param kwargs:
        :return:
        """
        kwargs = dict(kwargs)

        fill = kwargs.get('fill', self._canvas.fillcolor)
        if not isinstance(fill, Color):
            fill = Color(fill, mode='rgb', color_range=1)
        kwargs['fill'] = fill

        stroke = kwargs.get('stroke', self._canvas.strokecolor)
        if not isinstance(stroke, Color):
            stroke = Color(stroke, mode='rgb', color_range=1)
        kwargs['stroke'] = stroke

        kwargs['strokewidth'] = kwargs.get('strokewidth', self._canvas.strokewidth)
        inst = clazz(self, *args, **kwargs)
        return inst


    def EndClip(self, *args, **kwargs):
        return self._makeColorableInstance(EndClip, args, kwargs)
    def BezierPath(self, *args, **kwargs):
        return self._makeColorableInstance(BezierPath, args, kwargs)
    def ClippingPath(self, *args, **kwargs):
        return self._makeColorableInstance(ClippingPath, args, kwargs)
    def Rect(self, *args, **kwargs):
        return self._makeColorableInstance(Rect, args, kwargs)
    def Oval(self, *args, **kwargs):
        return self._makeColorableInstance(Oval, args, kwargs)
    def Ellipse(self, *args, **kwargs):
        return self._makeColorableInstance(Ellipse, args, kwargs)
    def Color(self, *args, **kwargs):
        return Color(*args, **kwargs)
    def Image(self, *args, **kwargs):
        return self._makeColorableInstance(Image, args, kwargs)
    def Text(self, *args, **kwargs):
        return self._makeColorableInstance(Text, args, kwargs)

    #### Variables

    def var(self, name, type, default=None, min=0, max=255, value=None, step=None):
        v = Variable(name, type, default=default, min=min, max=max, value=value, step=step)
        return self._addvar(v)


    #### Utility

    def color(self, *args):
        '''
        :param args: color in a supported format.

        :return: Color object containing the color.
        '''
        return self.Color(mode=self.color_mode, color_range=self.color_range, *args)

    choice = r.choice

    def random(self,v1=None, v2=None):
        # ipsis verbis from Nodebox
        if v1 is not None and v2 is None:
            if isinstance(v1, float):
                return r.random() * v1
            else:
                return int(r.random() * v1)
        elif v1 != None and v2 != None:
            if isinstance(v1, float) or isinstance(v2, float):
                start = min(v1, v2)
                end = max(v1, v2)
                return start + r.random() * (end-start)
            else:
                start = min(v1, v2)
                end = max(v1, v2) + 1
                return int(start + r.random() * (end-start))
        else: # No values means 0.0 -> 1.0
            return r.random()

    def grid(self, cols, rows, colSize=1, rowSize=1, shuffled=False):
        """Returns an iterator that contains coordinate tuples.
        The grid can be used to quickly create grid-like structures.
        A common way to use them is:
            for x, y in grid(10,10,12,12):
                rect(x,y, 10,10)
        """
        # Taken ipsis verbis from Nodebox
        from random import shuffle
        rowRange = range(int(rows))
        colRange = range(int(cols))
        if (shuffled):
            shuffle(rowRange)
            shuffle(colRange)
        for y in rowRange:
            for x in colRange:
                yield (x*colSize,y*rowSize)

    def files(self, path="*"):
        """Returns a list of files.
        You can use wildcards to specify which files to pick, e.g.
            f = files('*.gif')

        :param path: wildcard to use in file list.
        """
        # Taken ipsis verbis from Nodebox
        return glob(path)

    def snapshot(self,filename=None, surface=None, defer=None, autonumber=False):
        '''Save the contents of current surface into a file or cairo surface/context

        :param filename: Filename to save snapshot as, available formats include .png, .ps, .svg
        :param surface:  If specified will output snapshot to the supplied cairo surface.
        :param defer: If true, buffering/threading may be employed however output will not be immediate.
        :param autonumber: If true then a number will be appended to the filename.
        '''
        if autonumber:
            file_number=self._frame
        else:
            file_number=None
        if surface:
            if defer is None:
                defer=False
            self._canvas.snapshot(surface, defer)
            return
        elif filename is None:
            # If nothing specied, we can see if a filename is available
            script_file = self._namespace.get('__file__')
            if script_file:
                filename = os.path.splitext(script_file)[0] + '.svg'
                file_number=True

        if filename:
            if defer is None:
                defer=True
            self._canvas.snapshot(filename, defer=defer, file_number=file_number)
        else:
            raise ShoebotError('No image saved')
            
            

    def ximport(self, libName):
        '''
        Import nodebox libraries.

        The libraries get _ctx, which provides
        them with the nodebox API.

        :param libName: Library name to import
        '''
        # from Nodebox
        lib = __import__(libName)
        self._namespace[libName] = lib
        lib._ctx = self
        return lib


    #### Core functions

    def size(self, w = None, h = None):
        '''Set the canvas size

        Only the first call will actually be effective.

        :param w: Width
        :param h: height
        '''
        
        if not w:
            w = self._canvas.width
        if not h:
            h = self._canvas.height
        if not w and not h:
            return (self._canvas.width, self._canvas.height)

        # Updating in all these places seems a bit hacky
        w, h = self._canvas.set_size((w, h))
        self._namespace['WIDTH'] = w
        self._namespace['HEIGHT'] = h
        self.WIDTH = w  # Added to make evolution example work
        self.HEIGHT = h # Added to make evolution example work

    def speed(self, framerate=None):
        '''Set animation framerate.

        :param framerate: Frames per second to run bot.
        :return: Current framerate of animation.
        '''
        if framerate is not None:
            self._speed = framerate
            self._dynamic = True
        else:
            return self._speed

    @property
    def FRAME(self):
        return self._frame




from math import hypot, degrees, atan2, sqrt, pow, radians, sin, cos

def distance(x1, y1, x2, y2):
     return hypot(x2 - x1, y2 - y1)

def angle(x0, y0, x1, y1):
    # from nodebox1
    a = degrees( atan2(y1-y0, x1-x0) )
    return a

def coordinates(x0, y0, distance, angle):
    # from nodebox1
    x1 = x0 + cos(radians(angle)) * distance
    y1 = y0 + sin(radians(angle)) * distance
    return x1, y1

def reflect(x0, y0, x1, y1, d=1.0, a=180):
    # from nodebox1
    d *= distance(x0, y0, x1, y1)
    a += angle(x0, y0, x1, y1)
    x, y = coordinates(x0, y0, d, a)
    return x, y


from pysoundcard import Stream, continue_flag

import atexit
import collections
import time
import threading
import math
import numpy as np
import struct

from pysoundcard import devices
from fuzzywuzzy import fuzz


BUFF_LENGTH = 1024
CHANNELS = 2
NUM_SAMPLES = 512

def fft_bandpassfilter(data, fs, lowcut, highcut):
    """
    http://www.swharden.com/blog/2009-01-21-signal-filtering-with-python/#comment-16801
    """
    fft = np.fft.fft(data)
    n = len(data)
    timestep = 1.0 / fs
    freq = np.fft.fftfreq(n, d = timestep)
    bp = fft.copy()

    # Zero out fft coefficients
    #bp[10:-10] = 0

    # Normalise
    #bp *= real(fft.dot(fft))/real(bp.dot(bp))

    bp *= fft.dot(fft) / bp.dot(bp)
    
    # must multipy by 2 to get the correct amplitude
    ibp = 12 * np.fft.ifft(bp)
    return ibp

    #for i in range(len(bp)):
    #    if freq[i] >= highcut or freq[i] < lowcut:
    #        bp[i] = 0

    #ibp = np.fft.ifft(bp)
    #ibp = 2.0 * np.fft.ifft(bp)
    #print ibp
    #return ibp

def flatten_fft(scale = 1.0):
    """
    Produces a nicer graph, I'm not sure if this is correct
    """
    _len = len(audio.spectrogram)
    for i, v in enumerate(audio.spectrogram):
        yield scale * (i * v) / _len

def scaled_fft(fft, scale = 1.0):
    """
    Produces a nicer graph, I'm not sure if this is correct
    """
    data = np.zeros(len(fft))
    for i, v in enumerate(fft):
         data[i] = scale * (i * v) / NUM_SAMPLES
         
    return data

def triple(spectrogram):
    #c = spectrogram.copy()
    #c.resize(3, 255 / 3)
    #return c
    bass = spectrogram[0:85]
    mid = spectrogram[85:-85]
    treble = spectrogram[-85:-1]

    return bass, mid, treble

def fuzzydevices(match='', min_ratio=30):
    device_ratios = []
    for device in devices():
        ratio = fuzz.partial_ratio(match, device['name'])
        if ratio > min_ratio:
            device_ratios.append((ratio, device))

    for ratio, device in sorted(device_ratios, key=lambda ratio_device: (ratio_device[0])):
        yield device

def firstfuzzydevice(match=''):
    devices = list(fuzzydevices(match, 0))
    return devices[0]

class AudioException(Exception):
    def __init__(self, msg):
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg

class AudioThread(threading.Thread):
    def __init__(self):
        # TODO - This could be a bit less dumb, and only run
        #        bandpass (or other) filters if they are actually used
        super(self.__class__, self).__init__()
                
        self.daemon = True

        self.streams = {}
        self.running = False
        
        self._spectrogram = np.zeros(NUM_SAMPLES)
        self._bandpassed = np.zeros(NUM_SAMPLES)

    def settings(self, **kwargs):
        if self.running:
            raise AudioException('Audio is already running')

    def run(self):
        with Stream(sample_rate=44100, block_length=16) as s:
            while self.running:
                vec = s.read(NUM_SAMPLES)

                # Downsample to mono
                mono_vec = vec.sum(-1) / float(s.input_channels)

                self._spectrogram = np.fft.fft(mono_vec)

                self._bandpassed = fft_bandpassfilter(self.spectrogram, 44010, 100, 20)


    def autostart(self):
        if not self.running:
            self.running = True
            self.start()
            atexit.register(self.quit)

    def quit(self):
        """
        Shutdown the audio thread
        """        
        if self.running:
            self.running = False
            self.join()

    @property
    def spectrogram(self):
        self.autostart()
        return self._spectrogram

    @property
    def bandpassed(self):
        self.autostart()
        return self._bandpassed

audio = AudioThread()




import decimal
from unittest import TestCase

import simplejson as S

class TestDecode(TestCase):
    def test_decimal(self):
        rval = S.loads('1.1', parse_float=decimal.Decimal)
        self.assert_(isinstance(rval, decimal.Decimal))
        self.assertEquals(rval, decimal.Decimal('1.1'))

    def test_float(self):
        rval = S.loads('1', parse_int=float)
        self.assert_(isinstance(rval, float))
        self.assertEquals(rval, 1.0)



from unittest import TestCase

import simplejson as S
import textwrap

class TestIndent(TestCase):
    def test_indent(self):
        h = [['blorpie'], ['whoops'], [], 'd-shtaeou', 'd-nthiouh', 'i-vhbjkhnth',
             {'nifty': 87}, {'field': 'yes', 'morefield': False} ]

        expect = textwrap.dedent("""\
        [
          [
            "blorpie"
          ],
          [
            "whoops"
          ],
          [],
          "d-shtaeou",
          "d-nthiouh",
          "i-vhbjkhnth",
          {
            "nifty": 87
          },
          {
            "field": "yes",
            "morefield": false
          }
        ]""")


        d1 = S.dumps(h)
        d2 = S.dumps(h, indent=2, sort_keys=True, separators=(',', ': '))

        h1 = S.loads(d1)
        h2 = S.loads(d2)

        self.assertEquals(h1, h)
        self.assertEquals(h2, h)
        self.assertEquals(d2, expect)



# from cairocffi

import ctypes
import cairo  # pycairo
import cairocffi

pycairo = ctypes.PyDLL(cairo._cairo.__file__)
pycairo.PycairoContext_FromContext.restype = ctypes.c_void_p
pycairo.PycairoContext_FromContext.argtypes = 3 * [ctypes.c_void_p]
ctypes.pythonapi.PyList_Append.argtypes = 2 * [ctypes.c_void_p]


def _UNSAFE_cairocffi_context_to_pycairo(cairocffi_context):
    # Sanity check. Continuing with another type would probably segfault.
    if not isinstance(cairocffi_context, cairocffi.Context):
        raise TypeError('Expected a cairocffi.Context, got %r'
                        % cairocffi_context)

    # Create a reference for PycairoContext_FromContext to take ownership of.
    cairocffi.cairo.cairo_reference(cairocffi_context._pointer)
    # Casting the pointer to uintptr_t (the integer type as wide as a pointer)
    # gets the context's integer address.
    # On CPython id(cairo.Context) gives the address to the Context type,
    # as expected by PycairoContext_FromContext.
    address = pycairo.PycairoContext_FromContext(
        int(cairocffi.ffi.cast('uintptr_t', cairocffi_context._pointer)),
        id(cairo.Context),
        None)
    assert address
    # This trick uses Python's C API
    # to get a reference to a Python object from its address.
    temp_list = []
    assert ctypes.pythonapi.PyList_Append(id(temp_list), address) == 0
    return temp_list[0]


def _UNSAFE_pycairo_context_to_cairocffi(pycairo_context):
    # Sanity check. Continuing with another type would probably segfault.
    if not isinstance(pycairo_context, cairo.Context):
        raise TypeError('Expected a cairo.Context, got %r' % pycairo_context)

    # On CPython, id() gives the memory address of a Python object.
    # pycairo implements Context as a C struct:
    #     typedef struct {
    #         PyObject_HEAD
    #         cairo_t *ctx;
    #         PyObject *base;
    #     } PycairoContext;
    # Still on CPython, object.__basicsize__ is the size of PyObject_HEAD,
    # ie. the offset to the ctx field.
    # ffi.cast() converts the integer address to a cairo_t** pointer.
    # [0] dereferences that pointer, ie. read the ctx field.
    # The result is a cairo_t* pointer that cairocffi can use.
    return cairocffi.Context._from_pointer(
        cairocffi.ffi.cast('cairo_t **',
                           id(pycairo_context) + object.__basicsize__)[0],
        incref=True)




from random import random
from math import pi, sin, cos
from math import sqrt

class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

##### GRAPH LAYOUT ###################################################################################

class layout(object):
    
    """ Graph visualizer that calculates relative node positions.
    """
    
    def __init__(self, graph, iterations=1000):
        
        self.type = None
        self.graph = graph
        self.i = 0
        self.n = iterations
        
        self.__bounds = None

    def copy(self, graph):
        
        """ Returns a copy of the layout for the given graph.
        """
        
        l = self.__class__(graph, self.n)
        l.i = 0
        return l

    def prepare(self):
        
        for n in self.graph.nodes:
            n.vx = 0
            n.vy = 0
            n.force = Point(0,0)    
    
    def _bounds(self):
        
        min = Point(float("inf"), float("inf"))
        max = Point(float("-inf"), float("-inf"))
        for n in self.graph.nodes:
            if (n.vx < min.x): min.x = n.vx
            if (n.vy < min.y): min.y = n.vy
            if (n.vx > max.x): max.x = n.vx
            if (n.vy > max.y): max.y = n.vy
      
        return (min, max)

    bounds = property(_bounds)
    
    def _get_done(self):
        
        if self.i >= self.n: 
            return True
        return False
        
    done = property(_get_done)
    
    def iterate(self):
        
        self.i += 1
        return self.done
        
    def solve(self):
        
        while not self.done: 
            self.iterate()
            
    def reset(self):
        
        self.i = 0
        
    def refresh(self):
        
        self.i = self.n / 2

##### GRAPH CIRCLE LAYOUT ############################################################################

class circle_layout(layout):
    
    """ Simple layout with nodes arranged on one or more circles.
    """
    
    def __init__(self, graph, iterations=100):
    
        layout.__init__(self, graph, iterations)
        self.type = "circle"    
        
        self.r = 8    # outer circle radius
        self.c = 2    # number of circles
        self.a = pi/2 # starting angle
    
    def _get_orbits(self): return self.c
    def _set_orbits(self, v): self.c = v
    orbits = property(_get_orbits, _set_orbits)
    
    def copy(self, graph):
        
        l = layout.copy(self, graph)
        l.r = self.r
        return l
    
    def iterate(self):
        
        if len(self.graph.nodes) == 1: return
        
        # Nodes are sorted by betweenness centrality.
        # Node with a high centrality are on the inner circles.
        # There are logarithmically more nodes on the outer shells.
        circles = []
        nodes = self.graph.nodes_by_traffic(treshold=-1)
        for i in range(self.c):
            t = 1.0 / (self.c-i)**2
            slice = int(t * len(self.graph.nodes))
            slice = max(1, slice)
            circles.append(nodes[:slice])
            nodes = nodes[slice:]

        node_radius = self.graph.nodes[0].__class__(None).r

        i = 0
        for circle in circles:
            i += 1

            # Circle radii expand each iteration.
            # Inner circles have a smaller radius.
            r = self.r * sin(pi/2 * float(self.i) / self.n)
            r *= float(i)/len(circles)
            
            # Calculate circle circumference. 
            # Node diameter / circumference determine how many nodes fit on the shell.
            C = self.r*self.graph.d * 2*pi * float(i)/len(circles)
            s = node_radius*2 / C * 2
            
            a = self.a
            t = min(2*pi*s, 2*pi/len(circle))
            for n in circle:
                n.vx = r * cos(a)
                n.vy = r * sin(a)
                a += t

        #r = self.r * sin(float(self.i) / self.n * pi/2)
        #a = 0
        #i = pi*2 / len(self.graph.nodes)
        #for n in self.graph.nodes:
        #    n.vx = r * cos(a)
        #    n.vy = r * sin(a)
        #    a += i 
            
        layout.iterate(self)
        
    def solve(self):
        
        self.i = self.n
        self.iterate()

##### GRAPH SPRING LAYOUT ############################################################################

class spring_layout(layout):
    
    """ A force-based layout in which edges are regarded as springs.
    http://snipplr.com/view/1950/graph-javascript-framework-version-001/
    """
    
    def __init__(self, graph, iterations=1000):
        
        layout.__init__(self, graph, iterations)    
        self.type = "spring"

        self.k = 2    # force strength
        self.m = 0.01 # force multiplier
        self.w = 15   # edge weight multiplier
        self.d = 0.5  # maximum vertex movement
        self.r = 15   # maximum repulsive force radius
    
    def tweak(self, k=2, m=0.01, w=15, d=0.5, r=15):
        self.k = k
        self.m = m,
        self.w = w
        self.d = d
        self.r = r
    
    def _get_force(self): return self.m
    def _set_force(self, v): self.m = v
    force = property(_get_force, _set_force)

    def _get_repulsion(self): return self.r
    def _set_repulsion(self, v): self.r = v
    repulsion = property(_get_repulsion, _set_repulsion)
    
    def copy(self, graph):
        
        l = layout.copy(self, graph)
        l.k, l.m, l.d, l.r = self.k, self.m, self.d, self.r
        return l
    
    def iterate(self):
        
        # Forces on all nodes due to node-node repulsions.
        for i in range(len(self.graph.nodes)):
            n1 = self.graph.nodes[i]
            for j in range(i+1, len(self.graph.nodes)):
                n2 = self.graph.nodes[j]             
                self._repulse(n1, n2)

        # Forces on nodes due to edge attractions.
        for e in self.graph.edges:
            self._attract(e.node1, e.node2, self.w*e.weight, 1.0/e.length)
            
        # Move by given force.
        for n in self.graph.nodes:
            vx = max(-self.d, min(self.m * n.force.x, self.d))
            vy = max(-self.d, min(self.m * n.force.y, self.d))
            n.vx += vx
            n.vy += vy
            n.force.x = 0
            n.force.y = 0
        
        return layout.iterate(self)
    
    def _distance(self, n1, n2):
        
        dx = n2.vx - n1.vx
        dy = n2.vy - n1.vy
        d2 = dx**2 + dy**2       

        if d2 < 0.01:
            dx = random()*0.1 + 0.1
            dy = random()*0.1 + 0.1
            d2 = dx**2 + dy**2
            
        d = sqrt(d2)
        
        return dx, dy, d
    
    def _repulse(self, n1, n2):
        
        dx, dy, d = self._distance(n1, n2)
        
        if d < self.r:
            f = self.k**2 / d**2
            n2.force.x += f * dx
            n2.force.y += f * dy
            n1.force.x -= f * dx
            n1.force.y -= f * dy
        
    def _attract(self, n1, n2, k=0, length=1.0):
        
        dx, dy, d = self._distance(n1, n2)
        d = min(d, self.r)
        
        # Take the edge's weight (k) into account.
        f = (d**2 - self.k**2) / self.k * length
        f *= k * 0.5 + 1
        f /= d
        
        n2.force.x -= f * dx
        n2.force.y -= f * dy
        n1.force.x += f * dx
        n1.force.y += f * dy


# daisylib example library.
#
# To use this run the companion bot
# 'using_libraries.bot'
#
# Libraries using ximport get a variable called '_ctx'
# which they can use Shoebot/Nodebox commands on.
#
#
# TODO - If we create a 'Shoebot' grammar it should 
# provide a better mechanism than '_ctx'
#

from math import sin

def drawdaisy(x, y, color='#fefefe'):
    """
    Draw a daisy at x, y
    """
    # save location, size etc
    _ctx.push()
    
    # save fill and stroke
    _fill =_ctx.fill()
    _stroke = _ctx.stroke()

    sc = (1.0 / _ctx.HEIGHT) * float(y * 0.5) * 4.0

    # draw stalk
    _ctx.strokewidth(sc * 2.0)
    _ctx.stroke('#3B240B')
    
    _ctx.line(x + (sin(x * 0.1) * 10.0), y + 80, x + sin(_ctx.FRAME * 0.1), y)

    # draw flower
    _ctx.translate(-20, 0)
    _ctx.scale(sc)

    # draw petals
    _ctx.fill(color)
    _ctx.nostroke()
    for angle in xrange(0, 360, 45):
        _ctx.rotate(degrees=45)
        _ctx.rect(x, y, 40, 8, 1)

    # draw centre
    _ctx.fill('#F7FE2E')
    _ctx.ellipse(x + 15, y, 10, 10)

    # restore fill and stroke
    _ctx.fill(_fill)
    _ctx.stroke(_stroke)

    # restore location, size etc
    _ctx.pop()




#!/usr/bin/env python2

# This file is part of Shoebot.
# Copyright (C) 2007-2009 the Shoebot authors
# See the COPYING file for the full license text.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#   The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
#   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
#   EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
""" Shoebot console runner """

import locale, gettext
import argparse
import shlex
import sys

from __init__ import run

DEFAULT_SCRIPT = '/usr/share/shoebot/examples/primitives.py'
DEFAULT_WINDOW_SCRIPT = '/usr/share/shoebot/examples/socketcontrol2.py'
DEFAULT_OUTPUTFILE = 'output.png'
DEFAULT_SERVERPORT = 7777

OUTPUT_EXTENSIONS = ('.png', '.svg', '.ps', '.pdf')
APP = 'shoebot'
DIR = sys.prefix + '/share/shoebot/locale'

locale.setlocale(locale.LC_ALL, '')
gettext.bindtextdomain(APP, DIR)
#gettext.bindtextdomain(APP)
gettext.textdomain(APP)
_ = gettext.gettext

NODEBOX = 'nodebox'
DRAWBOT = 'drawbot'


def json_arg(s):
    #s = s.strip()
    if s.startswith("{"):
        # enclose in quotes
        s = "'%s'" % s

    print '...'
    print '>>%s<<' % s
    try:
        import json
        d = json.loads(s, parse_float=True, parse_int=True)
        return d
    except Exception as e:
        error(_('Error parsing JSON, remember single quotes OUTSIDE, double QUOTES inside.'))
        raise e


def error(message):
    '''Prints an error message, the help message and quits'''
    global parser
    print (_("Error: ") + message)
    print ()
    parser.print_help()
    sys.exit()


def warn(message):
    '''Print a warning message'''
    print (_("Warning: ") + message)


def main():
    global parser

    # use ArgumentParser to interpret commandline options
    parser = argparse.ArgumentParser(_("usage: sbot [options] inputfile.bot [args]"))
    parser.add_argument("script")
    parser.add_argument("-o",
                    "--outputfile",
                    dest="outputfile",
                    help=_("run script and output to FILE (accepts .svg, .ps, .pdf and .png extensions)"),
                    metavar="FILE")
    parser.add_argument("-w",
                    "--window",
                    action="store_true",
                    dest="window",
                    default=False,
                    help=_("run script in a GTK window")
                    )
    parser.add_argument("-f",
                    "--fullscreen",
                    action="store_true",
                    dest="fullscreen",
                    default=False,
                    help=_("run in fullscreen mode")
                    )
    parser.add_argument("-t",
                    "--title",
                    action="store",
                    dest="title",
                    default=None,
                    help=_("Set window title")
                    )
    parser.add_argument("-s",
                    "--socketserver",
                    action="store_true",
                    dest="socketserver",
                    default=False,
                    help=_("run a socket server for external control (will run the script in windowed mode)"))
    parser.add_argument("-dv",
                    "--disable-vars",
                    action="store_true",
                    dest="disable_vars",
                    default=False,
                    help=_("disable the variables pane when in windowed mode."))
    parser.add_argument("-dt",
                    "--disable-background-thread",
                    action="store_true",
                    dest="disable_background_thread",
                    default=False,
                    help=_("disable running bot code in background thread."))
    parser.add_argument("-p",
                    "--serverport",
                    type=int,
                    dest="serverport",
                    default=DEFAULT_SERVERPORT,
                    help=_("set socketserver port to listen for connections (default is 7777)"))
    parser.add_argument("-r",
                    "--repeat",
                    type=int,
                    dest="repeat",
                    default=False,
                    help=_("set number of iteration, multiple images will be produced"))
    parser.add_argument("-g",
                    "--grammar",
                    dest="grammar",
                    default=NODEBOX,
                    help=_("Select the bot grammar 'nodebox' (default) or 'drawbot' languages"),
                    )
    parser.add_argument("-c",
                    "--close",
                    action="store_true",
                    dest="close",
                    default=False,
                    help=_("Close window after running bot (use with -r for benchmarking)"),
                    )
    parser.add_argument("-v",
                    "--vars",
                    dest="vars",
                    default=False,
                    help=_("Initial variables, in JSON (Note: Single quotes OUTSIDE, double INSIDE) --vars='{\"variable1\": 1}'"),
                    )
    parser.add_argument("-ns",
                    "--namespace",
                    dest="namespace",
                    default=None,
                    help=_("Initial namespace, in JSON (Note: Single quotes OUTSIDE, double INSIDE) --namespace='{\"variable1\": 1}'"),
                    )
    parser.add_argument("-l",
                    "--l",
                    dest="shell",
                    action="store_true",
                    default=False,
                    help=_("Simple shell - for IDE interaction"),
                    )
    parser.add_argument("-a",
                    "--args",
                    dest="script_args",
                    help=_("Pass to the bot"),
                    )
    parser.add_argument("-V",
                    "--verbose",
                    action="store_true",
                    dest="verbose",
                    default=False,
                    help=_("Show internal shoebot error information in traceback"),
                    )
    parser.add_argument('script_args', nargs='?')

    # get argparse arguments and check for sanity
    args, extra = parser.parse_known_args()

    if not args.script and not args.window:
        error(_('Please specify an input script!\n (check /usr/share/shoebot/examples/ for example scripts)'))

    vars = None
    if args.vars:
        vars = json_arg(args.vars)
        # try:
        #     import json
        #     vars = json.loads(args.vars)
        # except Exception as e:
        #     error(_('Error parsing JSON, remember single quotes OUTSIDE, double QUOTES inside.'))
        #     raise e

    namespace = None
    if args.namespace:
        namespace = json_arg(args.namespace)
        # try:
        #     import json
        #     namespace = json.loads(args.namespace)
        # except Exception as e:
        #     error(_('Error parsing JSON, remember single quotes OUTSIDE, double QUOTES inside.'))
        #     raise e

    run(src = args.script,
        grammar = args.grammar,
        outputfile = args.outputfile,
        iterations = args.repeat or None,
        window = args.window or args.socketserver,
        fullscreen = args.fullscreen,
        title = args.title,
        close_window = args.close,
        server=args.socketserver,
        port=args.serverport,
        show_vars = args.window and args.disable_vars == False,
        vars = vars or None,
        namespace = namespace,
        run_shell = args.shell,
        args = shlex.split(args.script_args or ""),
        verbose = args.verbose,
        background_thread=not args.disable_background_thread,
        )

if __name__ == '__main__':
    main()


import url
import soap
import re
from cache import Cache
import pickle

def clear_cache():
    Cache("urbandictionary").clear()

class UrbanDictionaryDefinition:
    
    def __init__(self, word, url, description, example, author):

        self.word = word
        self.url = url
        self.description = description
        self.example = example
        self.author = author
        self._parse()
    
    def _parse(self):
        
        """ Strips links from the definition and gathers them in a links property.
        """
        
        p1 = "\[.*?\](.*?)\[\/.*?\]"
        p2 = "\[(.*?)\]"
        self.links = []
        for p in (p1,p2):
            for link in re.findall(p, self.description):
                self.links.append(link)
            self.description = re.sub(p, "\\1", self.description)
            
        self.description = self.description.strip()
    
    def __str__(self):
        
        return self.description

class UrbanDictionaryError(Exception):
    pass

class UrbanDictionary(list):
    
    def __init__(self, q, cached=True):
        
        url = "http://api.urbandictionary.com/soap?wsdl"
        key = "91cf66fb7f14bbf7fb59c7cf5e22155f"

        # Live connect for uncached queries 
        # or queries we do not have in cache.
        cache = Cache("urbandictionary", ".pickle")
        if not cached or not cache.exists(q):
            server = soap.SOAPProxy(url)
            try:
                definitions = server.lookup(key, q)
            except Exception, soap.faultType:
                raise UrbanDictionaryError, "the API is no longer supported"
            data = []
            for item in definitions:
                ubd = UrbanDictionaryDefinition(
                    item.word, item.url, item.definition, item.example, item.author
                )
                self.append(ubd)
                data.append( [item.word, item.word, item.definition, item.example, item.author] )
            # Cache a pickled version of the response.
            if cached:
                data = pickle.dumps(data)
                cache.write(q, data)
        
        # For cached queries,
        # unpack the pickled version in the cache.
        else:
            definitions = cache.read(q)
            definitions = pickle.loads(definitions)
            for item in definitions:
                ubd = UrbanDictionaryDefinition(
                    item[0], item[1], item[2], item[3], item[4]
                )
                self.append(ubd)
            
def search(q, cached=True):
    return UrbanDictionary(q, cached)


class DrawQueueSink(object):
    
    '''
    DrawQueueSink, creates parameters for use by the draw queue.
    (the render_context).

    The render context is a set of platform sepecific
    parameters used by implementations of the drawqueue,
    canvas, and sink.
    '''

    def set_bot(self, bot):
        self.bot = bot

    def render(self, size, frame, drawqueue):
        '''
        Calls implmentation to get a render context,
        passes it to the drawqueues render function
        then calls self.rendering_finished
        '''
        r_context = self.create_rcontext(size, frame)
        drawqueue.render(r_context)
        return r_context

    def create_rcontext(self, size, frame):
        '''
        Returns a cairo context for drawing this
        frame of the bot
        '''
        pass
    
    def main_iteration(self):
        """
        Called from main loop, if your sink needs to handle GUI events
        do it here.

        :return:
        """
        pass



### CREDITS ##########################################################################################

# Copyright (c) 2007 Tom De Smedt.
# See LICENSE.txt for details.

__author__    = "Tom De Smedt"
__version__   = "1.9.2"
__copyright__ = "Copyright (c) 2007 Tom De Smedt"
__license__   = "MIT"

import sys, os
sys.path.append(os.path.dirname(__file__))

from sqlite3 import dbapi2 as sqlite



### DATABASE #########################################################################################

class Database:
    
    def __init__(self):
        
        self._name = None
        self._tables = [] # All table names.
        self._indices = []
        
        self._commit = 0
        self._i = 0

    def connect(self, name):
        
        """Generic database.
        
        Opens the SQLite database with the given name.
        The .db extension is automatically appended to the name.
        For each table in the database an attribute is created,
        and assigned a Table object.
        
        You can do: database.table or database[table].
        
        """
        
        self._name = name.rstrip(".db")
        self._con = sqlite.connect(self._name + ".db")
        self._cur = self._con.cursor()
 
        self._tables = []
        self._cur.execute("select name from sqlite_master where type='table'")
        for r in self._cur: self._tables.append(r[0])
        
        self._indices = []
        self._cur.execute("select name from sqlite_master where type='index'")
        for r in self._cur: self._indices.append(r[0])
        
        for t in self._tables:
            self._cur.execute("pragma table_info("+t+")")
            fields = []
            key = ""
            for r in self._cur:
                fields.append(r[1])
                if r[2] == "integer": key = r[1]
            setattr(self, t, Table(self, t, key, fields))
        
    def create(self, name, overwrite=True):
        
        """Creates an SQLite database file.
        
        Creates an SQLite database with the given name.
        The .box file extension is added automatically.
        Overwrites any existing database by default.
        
        """
        
        self._name = name.rstrip(".db")
        from os import unlink
        if overwrite: 
            try: unlink(self._name + ".db")
            except: pass       
        self._con = sqlite.connect(self._name + ".db")
        self._cur = self._con.cursor()        
    
    def __len__(self):
        return len(self._tables)
    
    def __getitem__(self, name):
        if isinstance(name, int): name = self._tables[name]
        return getattr(self, name)
        
    # Deprecated, use for-loop and Database[table_name] instead.
    def tables(self): return self._tables
    table = __getitem__
    
    def create_table(self, name, fields=[], key="id"):
        
        """Creates a new table.
        
        Creates a table with the given name,
        containing the list of given fields.
        Since SQLite uses manifest typing, no data type need be supplied.
        The primary key is "id" by default,
        an integer that can be set or otherwise autoincrements.
        
        """
        
        for f in fields: 
            if f == key: fields.remove(key)
        sql  = "create table "+name+" "
        sql += "("+key+" integer primary key"
        for f in fields: sql += ", "+f+" varchar(255)"
        sql += ")"
        self._cur.execute(sql)
        self._con.commit()
        self.index(name, key, unique=True)
        self.connect(self._name)

    # Deprecated, use create_table() instead.
    append = create_table
    
    def create_index(self, table, field, unique=False, ascending=True):
        
        """Creates a table index.
        
        Creates an index on the given table,
        on the given field with unique values enforced or not,
        in ascending or descending order.
        
        """
        
        if unique: u = "unique "
        else: u = ""
        if ascending: a = "asc"
        else: a = "desc"
        sql  = "create "+u+"index index_"+table+"_"+field+" "
        sql += "on "+table+"("+field+" "+a+")"
        self._cur.execute(sql)
        self._con.commit()
    
    # Deprecated, use create_index() instead.
    index = create_index
        
    def commit(self, each=0):
        
        """Sets the commit frequency.
        
        Modifications to the database,
        e.g. row insertions are commited in batch,
        specified by the given number.
        A number that is reasonably high allows for faster transactions.
        Commits anything still pending.
        
        """
        
        self._commit = each
        self._con.commit()
        
    def close(self):
        
        """Commits any pending transactions and closes the database.
        """
        
        self._con.commit()
        self._cur.close()
        self._con.close()
        
    def sql(self, sql):
        
        """ Executes a raw SQL statement on the database.
        """
        
        self._cur.execute(sql)
        if sql.lower().find("select") >= 0:
            matches = []
            for r in self._cur: matches.append(r)
            return matches
            
    def dump(self, ext=".xml"):
        
        """ Dumps the data in the tables into another format (like XML).
        """
        
        self._con.commit()
        if ext == ".xml":
            return self._dump_xml()
    
    def _dump_xml(self):
        
        data  = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
        data += "<database name=\""+self._name+"\">\n"
        for name in self._indices:
            s = name.split("_")
            data += "<index name=\""+name+"\" table=\""+s[0]+"\" field=\""+s[1]+"\">\n"
            
        for table in self:
            data += "<table name=\""+table._name+"\" key=\""+table._key+"\">\n"
            for row in table.all():
                data += "\t<row id=\""+unicode(row[table._fields.index(table._key)])+"\">\n"
                i = 0
                for field in table._fields:
                    r = unicode(row[i])
                    if row[i] == None: r = ""
                    data += "\t\t<"+field+">"+r+"</"+field+">\n"
                    i += 1
                data += "\t</row>\n"
            data += "</table>\n"
            
        data += "</database>"
        
        f = open(self._name+".xml", "w")
        f.write(data)        
        f.close()
        
        return data

    # Deprecated, use dump() instead.
    xml = dump

### TABLE ############################################################################################

class Table:

    def __init__(self, db, name, key, fields):
        
        """Generic table.
        
        Constructs a table with the given name, primary key and columns.
        Each of the column names becomes the name of a table method
        that fetches rows from the table.
        
        For example, a table with an id field has the following method:
        table.id(query, operator="=")
        
        """
    
        self._db = db
        self._name = name
        self._key = key
        self._fields = fields
        
        for f in self._fields:
            import new
            im = lambda t, q, operator="=", fields="*", _field=f: t.find(q, operator, fields, _field)
            setattr(self, f, new.instancemethod(im, self, None))

    def _get_name(self): return self._name
    name = property(_get_name)

    def __len__(self):
        
        """ The row count of the table. This should be optimized.
        """
        
        sql = "select "+self._key+" from "+self._name
        self._db._cur.execute(sql)
        i = 0
        for r in self._db._cur: i += 1
        return i

    def find(self, q, operator="=", fields="*", key=None):
        
        """A simple SQL SELECT query.
        
        Retrieves all rows from the table 
        where the given query value is found in the given column (primary key if None).
        A different comparison operator (e.g. >, <, like) can be set.
        The wildcard character is * and automatically sets the operator to "like".
        Optionally, the fields argument can be a list of column names to select.
        Returns a list of row tuples containing fields.
        
        """
        
        if key == None: key = self._key
        if fields != "*": fields = ", ".join(fields)
        try: q = unicode(q)
        except: pass
        if q != "*" and (q[0] == "*" or q[-1] == "*"):
            if q[0]  == "*": q = "%"+q.lstrip("*")
            if q[-1] == "*": q = q.rstrip("*")+"%"
            operator = "like"
        
        if q != "*":
            sql = "select "+fields+" from "+self._name+" where "+key+" "+operator+" ?"
            self._db._cur.execute(sql, (q,))
        else:
            sql = "select "+fields+" from "+self._name
            self._db._cur.execute(sql)
        
        # You need to watch out here when bad unicode data 
        # has entered the database: pysqlite will throw an OperationalError.
        # http://lists.initd.org/pipermail/pysqlite/2006-April/000488.html
        matches = []
        for r in self._db._cur: matches.append(r)
        return matches

    def all(self):
        
        """Returns all the rows in the table.
        """
        
        return self.find("*")
        
    def fields(self):
        
        """Returns the column names.
        
        Returns the name of each column in the database,
        in the same order row fields are returned from find().
        
        """
        
        return self._fields
    
    def append(self, *args, **kw):
        
        """Adds a new row to a table.
        
        Adds a row to the given table.
        The column names and their corresponding values
        must either be supplied as a dictionary of {fields:values},
        or a series of keyword arguments of field=value style.
        
        """

        if args and kw: 
            return
        if args and type(args[0]) == dict:
            fields = [k for k in args[0]]
            v = [args[0][k] for k in args[0]]
        if kw:
            fields = [k for k in kw]
            v = [kw[k] for k in kw]
        
        q = ", ".join(["?" for x in fields])
        sql  = "insert into "+self._name+" ("+", ".join(fields)+") "
        sql += "values ("+q+")"
        self._db._cur.execute(sql, v)
        self._db._i += 1
        if self._db._i >= self._db._commit:
            self._db._i = 0
            self._db._con.commit()
            
    def edit(self, id, *args, **kw):
        
        """ Edits the row with given id.
        """
        
        if args and kw: 
            return
        if args and type(args[0]) == dict:
            fields = [k for k in args[0]]
            v = [args[0][k] for k in args[0]]
        if kw:
            fields = [k for k in kw]
            v = [kw[k] for k in kw]
        
        sql  = "update "+self._name+" set "+"=?, ".join(fields)+"=? where "+self._key+"="+unicode(id)
        self._db._cur.execute(sql, v)
        self._db._i += 1
        if self._db._i >= self._db._commit:
            self._db._i = 0
            self._db._con.commit()        
            
    def remove(self, id, operator="=", key=None):

        """ Deletes the row with given id.
        """

        if key == None: key = self._key
        try: id = unicode(id)
        except: pass        
        sql = "delete from "+self._name+" where "+key+" "+operator+" ?"
        self._db._cur.execute(sql, (id,))

######################################################################################################

def create(name, overwrite=True):
    
    db = Database()
    db.create(name, overwrite)
    return db
    
def connect(name):
    
    db = Database()
    db.connect(name)
    return db



#
# transforms.py from old shoebot
#

import sys
import cairocffi as cairo
from math import sin, cos
from bezier import BezierPath
                         
TRANSFORMS = ['translate', 'scale', 'rotate', 'skew', 'push', 'pop']
CENTER = 'center'
CORNER = 'corner'
CORNERS = "corners"

import locale, gettext
APP = 'shoebot'
DIR = sys.prefix + '/share/shoebot/locale'
locale.setlocale(locale.LC_ALL, '')
gettext.bindtextdomain(APP, DIR)
#gettext.bindtextdomain(APP)
gettext.textdomain(APP)
_ = gettext.gettext


class Transform:
    '''
    This class represents a stack of transformations. Supported operations are
    translation, scaling, rotation and skewing.

    '''
    def __init__(self, transform=None):
        self.stack = []
        if transform is None:
            pass
        elif isinstance(transform, Transform):
            self.append(transform)
        elif isinstance(transform, (list, tuple)):
            matrix = tuple(transform)
            t = cairo.Matrix(*matrix)
            self.append(t)
        elif isinstance(transform, cairo.Matrix):
            self.append(transform)
        else:
            raise ValueError, _("Transform: Don't know how to handle transform %s.") % transform

    def translate(self, x, y):
        t = ('translate', x, y)
        self.stack.append(t)

    def scale(self, x, y):
        t = ('scale', x, y)
        self.stack.append(t)

    def rotate(self, a):
        t = ('rotate', a)
        self.stack.append(t)

    def skew(self, x, y):
        t = ('skew', x, y)
        self.stack.append(t)

    def push(self):
        t = ('push',)
        self.stack.append(t)

    def pop(self):
        t = ('pop',)
        self.stack.append(t)


    def append(self, t):
        if isinstance(t, Transform):
            for item in t.stack:
                self.stack.append(item)
        elif isinstance(t, cairo.Matrix):
            self.stack.append(t)
        else:
            raise ValueError(_("Transform: Can only append Transforms or Cairo matrices (got %s)") % (t))

    def prepend(self,t):
        if isinstance(t, Transform):
            newstack = []
            for item in t.stack:
                newstack.append(item)
            for item in self.stack:
                newstack.append(item)
            self.stack = newstack
        elif isinstance(t, cairo.Matrix):
            self.stack.insert(0,t)
        else:
            raise ValueError(_("Transform: Can only append Transforms or Cairo matrices (got %s)") % (t))

    def copy(self):
        return self.__class__(self)
    def __iter__(self):
        for value in self.stack:
            yield value
    ### calculates tranformation matrix
    def get_matrix_with_center(self,x,y,mode):
        m = cairo.Matrix()
        centerx =x
        centery = y
        m_archived = []
        
        for trans in self.stack: 
            if isinstance(trans, cairo.Matrix):
                # multiply matrix
                m *= trans
            elif isinstance(trans, tuple) and trans[0] in TRANSFORMS:
                # parse transform command
                cmd = trans[0]
                args = trans[1:]
                t = cairo.Matrix()
                
                if cmd == 'translate':                    
                    xt = args[0]
                    yt = args[1]
                    m.translate(xt,yt)
                elif cmd == 'rotate':
                    if mode == 'corner':                        
                        # apply existing transform to cornerpoint
                        deltax,deltay = m.transform_point(0,0)
                        a = args[0]
                        ct = cos(a)
                        st = sin(a)
                        m *= cairo.Matrix(ct, st, -st, ct,deltax-(ct*deltax)+(st*deltay),deltay-(st*deltax)-(ct*deltay)) 
                    elif mode == 'center':
                        # apply existing transform to centerpoint
                        deltax,deltay = m.transform_point(centerx,centery)
                        a = args[0]
                        ct = cos(a)
                        st = sin(a)
                        m *= cairo.Matrix(ct, st, -st, ct,deltax-(ct*deltax)+(st*deltay),deltay-(st*deltax)-(ct*deltay)) 
                elif cmd == 'scale':
                    if mode == 'corner':
                        t.scale(args[0], args[1])
                        m *= t
                    elif mode == 'center':
                        # apply existing transform to centerpoint
                        deltax,deltay = m.transform_point(centerx,centery)
                        x, y = args
                        m1 = cairo.Matrix()
                        m2 = cairo.Matrix()
                        m1.translate(-deltax, -deltay)
                        m2.translate(deltax, deltay)
                        m *= m1
                        m *= cairo.Matrix(x,0,0,y,0,0)
                        m *= m2

                elif cmd == 'skew':
                    if mode == 'corner':
                        x, y = args
                        ## TODO: x and y should be the tangent of an angle
                        t *= cairo.Matrix(1,0,x,1,0,0)
                        t *= cairo.Matrix(1,y,0,1,0,0)
                        m *= t
                    elif mode == 'center':
                        # apply existing transform to centerpoint
                        deltax,deltay = m.transform_point(centerx,centery)
                        x,y = args
                        m1 = cairo.Matrix()
                        m2 = cairo.Matrix()
                        m1.translate(-deltax, -deltay)
                        m2.translate(deltax, deltay)
                        t *= m
                        t *= m1
                        t *= cairo.Matrix(1,0,x,1,0,0)
                        t *= cairo.Matrix(1,y,0,1,0,0)
                        t *= m2
                        m = t
                elif cmd == 'push':
                    m_archived.append(m)
                elif cmd == 'pop':
                    m = m_archived.pop()

        return m        

    def get_matrix(self):
        '''Returns this transform's matrix. Its centerpoint is presumed to be
        (0,0), which is the Cairo default.'''
        return self.get_matrix_with_center(0,0)



    def transformBezierPath(self, path):
        # From nodebox 
        if isinstance(path, BezierPath):
            path = path.copy()
        else:
            raise ValueError("Can only transform BezierPaths")
         
        for point in path:
            print point   
        #path._nsBezierPath = self._nsAffineTransform.transformBezierPath_(path._nsBezierPath)
        return path


class TransformMixin(object):

    """Mixin class for transformation support.
    Adds the _transform and _transformmode attributes to the class."""

    def __init__(self):
        self._reset()

    def _reset(self):
        self._transform = Transform()
        self._transformmode = CENTER        

    def _get_transform(self):
        return self._transform
    def _set_transform(self, transform):
        self._transform = Transform(transform)
    transform = property(_get_transform, _set_transform)

    def _get_transformmode(self):
        return self._transformmode
    def _set_transformmode(self, mode):
        self._transformmode = mode
    transformmode = property(_get_transformmode, _set_transformmode)

    def reset(self):
        self._transform = Transform()
    def rotate(self, degrees=0, radians=0):
        self._transform.rotate(degrees,radians)
    def translate(self, x=0, y=0):
        self._transform.translate(x,y)
    def scale(self, x=1, y=None):
        self._transform.scale(x,y)
    def skew(self, x=0, y=0):
        self._transform.skew(x,y)





import unittest
import doctest

def additional_tests():
    import simplejson
    import simplejson.encoder
    import simplejson.decoder
    suite = unittest.TestSuite()
    for mod in (simplejson, simplejson.encoder, simplejson.decoder):
        suite.addTest(doctest.DocTestSuite(mod))
    return suite

def main():
    suite = additional_tests()
    runner = unittest.TextTestRunner()
    runner.run(suite)

if __name__ == '__main__':
    import os
    import sys
    sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    main()


# Shoebot Video Library 0.2
# Copyright: HVA - Hermanitos Verdes Architetti, 2009
# licenSe: LGPL

__author__    = "Francesco Fantoni"
__version__   = "0.2"
__copyright__ = "Copyright (c) 2009, HVA - Hermanitos Verdes Architetti"
__license__   = "lgpl"


import cairo, numpy
import opencv
from opencv import highgui as hg
import os

class SBVideoError(RuntimeError):
    pass

class Movie:
    
    def __init__(self, path, start=0, stop=None):
        self.path = path
        self.video = hg.cvCreateFileCapture(self.path)
        if self.video is None:
            raise SBVideoError("Could not open stream %s" % self.path)

        # these functions don't seem to work at present on my linux system
        
        #self.fps = hg.cvGetCaptureProperty(self.video, hg.CV_CAP_PROP_FPS)
        #self.n_of_frames = hg.cvGetCaptureProperty(self.video, hg.CV_CAP_PROP_FRAME_COUNT)
        #self.duration = self.n_of_frames/self.fps
        #self.width = hg.cvGetCaptureProperty(self.video, hg.CV_CAP_PROP_FRAME_WIDTH)
        #self.height = hg.cvGetCaptureProperty(self.video, hg.CV_CAP_PROP_FRAME_HEIGHT)
        hg.cvSetCaptureProperty(self.video, hg.CV_CAP_PROP_POS_FRAMES, start)
    
    def frame(self, t=None):
        frame = MovieFrame(src=self.video, time=t)
        return frame                


def movie(path, start=0, stop=None):
    return Movie(path, start, stop)


class Camera:
    
    def __init__(self, cam=0, width=None, height=None):
        self.path = cam
        self.video = hg.cvCreateCameraCapture(self.path)
        if width:
            hg.cvSetCaptureProperty(self.video, hg.CV_CAP_PROP_FRAME_WIDTH, width)
        if height:
            hg.cvSetCaptureProperty(self.video, hg.CV_CAP_PROP_FRAME_HEIGHT, height)
    
    def frame(self, t=None):
        frame = MovieFrame(src=self.video, time=None)
        return frame

def camera(cam=0, width=None, height=None):
    return Camera(cam, width, height)


class MovieFrame:

    def __init__(self, src="", time=None):

        self.src = src
        self.time = time
        if self.time:
            hg.cvSetCaptureProperty(self.src, hg.CV_CAP_PROP_POS_FRAMES, self.time)    
        self.iplimage = hg.cvQueryFrame(self.src)
        self.width = self.iplimage.width
        self.height = self.iplimage.height
        self.image = opencv.cvCreateImage(opencv.cvGetSize(self.iplimage),8, 4)
        opencv.cvCvtColor(self.iplimage,self.image,opencv.CV_BGR2BGRA)
        self.buffer = numpy.fromstring(self.image.imageData, dtype=numpy.uint32).astype(numpy.uint32)
        self.buffer.shape = (self.image.width, self.image.height)
        self.time = hg.cvGetCaptureProperty(self.src, hg.CV_CAP_PROP_POS_MSEC)
                
    def _data(self):
        return cairo.ImageSurface.create_for_data(self.buffer, cairo.FORMAT_RGB24, self.width, self.height, self.width*4)
        
    data = property(_data)
    
    def detectObject(self, classifier):
        self.grayscale = opencv.cvCreateImage(opencv.cvGetSize(self.iplimage), 8, 1)
        opencv.cvCvtColor(self.iplimage, self.grayscale, opencv.CV_BGR2GRAY)
        self.storage = opencv.cvCreateMemStorage(0)
        opencv.cvClearMemStorage(self.storage)
        opencv.cvEqualizeHist(self.grayscale, self.grayscale)
        
        try:
            self.cascade = opencv.cvLoadHaarClassifierCascade(os.path.join(os.path.dirname(__file__), classifier+".xml"),opencv.cvSize(1,1))
        except:
            raise AttributeError, "could not load classifier file"            
        
        self.objects = opencv.cvHaarDetectObjects(self.grayscale, self.cascade, self.storage, 1.2, 2, opencv.CV_HAAR_DO_CANNY_PRUNING, opencv.cvSize(50,50))
        
        return self.objects        
        
    def _faces(self):
        classifier = "haarcascade_frontalface_alt"
        return self.detectObject(classifier)
    faces = property(_faces)
    
class objecttest:
    
    def __init__(self, i):
        self.x=1*i
        self.y=1*i
        self.width=10*i
        self.height=10*i 
        
                              



### CREDITS ##########################################################################################

# Copyright (c) 2007 Frederik De Bleser.
# See LICENSE.txt for details.

__author__    = "Frederik De Bleser, Mark Meyer, Tom De Smedt"
__version__   = "1.9.3"
__copyright__ = "Copyright (c) 2007 Frederik De Bleser"
__license__   = "MIT"

### L-SYSTEM #########################################################################################

from shoebot.data import CORNER, CENTER
from sys import maxint

class LSystem(object):
    
    def __init__(self):
        
        self.angle = 20
        self.segmentlength = 40
        self.decrease = 0.7
        self.threshold = 3.0
        
        # The system's growth pattern.
        # Each rule can contain command symbols 
        # or a character referring to another rule.
        # This default ruleset draws a nice geometric tree.
        self.rules = {
        	"1" : "FFF[--2][-2][+2][++2]",
        	"2" : "FFF[--3][-3][+4][++4]",
        	"3" : "FFF[--5][-5][+5][++5]",
        	"4" : "FFF[--5][-5][+5][++5]",
        	"5" : "FFFFF[--5][++5]"   
        }
        self.root = "1"
        
        # You also define your own command symbols.
        self.commands = {}
        
        self._segments = 0
        self._duration = 0
        self._timed = False
        self.cost = 0.25
        
    def _get_d(self): return self.segmentlength
    def _set_d(self, v): self.segmentlength = v
    d = property(_get_d, _set_d)
		
    def _reset(self):
    	
		""" Resets the number of drawn segments and the duration.

		To calculate the number of segments or the total time needed,
		we need to recurse through LSystem._grow().
		Before that, the draw(), segments() and duration() command
		will reset both tally variables.    	
    	
		"""

		self._segments = 0
		self._duration = 0

    def _grow(self, generation, rule, angle, length, time=maxint, draw=True):

        """ Recurse through the system.
        
        When a segment is drawn, the LSsytem.segment() method will be called.
        You can customize this method to create your own visualizations.
        It takes an optional time parameter. 
        
        If you divide this parameter by LSsytem.duration() you get 
        a number between 0.0 and 1.0 you can use as an alpha value for example.
        
        The method also has an id parameter which is a unique number 
        between 0 and LSystem.segments.
        
        """

        if generation == 0: 
        	# We are at the bottom of the system so now we now the total time needed.
        	self._duration = 1 + maxint-time
        
        if length <= self.threshold: 
        	# Segment length has fallen below the threshold, stop recursing.
        	self._duration = 1 + maxint-time
        	return

        if rule in self.commands: 
        	# Custom command symbols:
        	# If the rule is a key in the LSsytem.commands dictionary,
        	# execute its value which is a function taking 6 parameters:
        	# lsystem, generation, rule, angle, length and time.
            self.commands[rule](self, generation, rule, angle, length, time)
        
        if draw:
            # Standard command symbols:
            # f signifies a move,
            # + and - rotate either left or right, | rotates 180 degrees,
            # [ and ] are for push() and pop(), e.g. offshoot branches,
            # < and > decrease or increases the segment length,
            # ( and ) decrease or increases the rotation angle.
	        if   rule == "f": _ctx.translate(0, -min(length, length*time))
	        elif rule == "-": _ctx.rotate(max(-angle, -angle*time))
	        elif rule == "+": _ctx.rotate(min(+angle, +angle*time))
	        elif rule == "|": _ctx.rotate(180)
	        elif rule == "[": _ctx.push()
	        elif rule == "]": _ctx.pop()

        if rule in self.rules \
        and generation > 0 \
        and time > 0:
        	# Recursion:
        	# Occurs when there is enough "life" (i.e. generation or time).
        	# Generation is decreased and segment length scaled down.
        	# Also, F symbols in the rule have a cost that depletes time.
            for cmd in self.rules[rule]:
            	# Modification command symbols:
                if   cmd == "F": time -= self.cost
                elif cmd == "!": angle = -angle
                elif cmd == "(": angle *= 1.1
                elif cmd == ")": angle *= 0.9
                elif cmd == "<": length *= 0.9
                elif cmd == ">": length *= 1.1
                self._grow(
                    generation-1,
                    cmd, 
                    angle, 
                    length*self.decrease, 
                    time,
                    draw
                )
        
        elif rule == "F" \
        or (rule in self.rules and self.rules[rule] == ""):
        	# Draw segment:
        	# If the rule is an F symbol or empty (e.g. in Penrose tiles).
        	# Segment length grows to its full size as time progresses.
            self._segments += 1
            if draw and time > 0:
            	length = min(length, length*time)
            	if self._timed:
            	    self.segment(length, generation, time, id=self._segments)
                else:
                    self.segment(length, generation, None, id=self._segments)
            	_ctx.translate(0, -length)
        
    def segment(self, length, generation, time=None, id=None):
    	_ctx.push()
    	_ctx.line(0, 0, 0, -length)
    	_ctx.scale(0.65)
    	_ctx.rect(-length/2, -length*3/2, length, length)
    	_ctx.pop()
       
    def draw(self, x, y, generation, time=None, ease=None):
    	
    	""" Draws a number of generations at the given position.
    	
    	The time parameter can be used to progress the system in an animatiom.
    	As time nears LSystem.duration(generation), more segments will be drawn.
    	
    	The ease parameter can be used to gradually increase the branching angle
    	as more segments are drawn.
    	
    	"""
        
        angle = self.angle
        if time is not None and ease:
        	angle = min(self.angle, self.angle * time / ease)

    	self._timed = True
    	if not time:
    		self._timed = False
    		time = maxint        

    	mode = _ctx.transform()
        _ctx.transform(CORNER)
        _ctx.push()
        _ctx.translate(x, y)
        self._reset()
        self._grow(generation, self.root, angle, self.d, time, draw=True)
        _ctx.pop()
        _ctx.transform(mode)
        
    def segments(self, generation, time=None):
    	
    	""" Returns the number of segments drawn for a number of generations.
    	
    	The number of segments that are drawn to the screen
    	depends of the number of generations and the amount of time.
    	Each F command has a cost that depletes time.
    	Segments will stop being drawn if generation reaches 0,
    	when there is no time left 
    	or when the segment length falls below LSystem.threshold.
    	
    	"""
    	
    	if not time: 
    		time = maxint
    	_ctx.push()
        self._reset()
        self._grow(generation, self.root, self.angle, self.d, time, draw=False)
        _ctx.pop()
        return self._segments

    def duration(self, generation):
    	
    	""" Returns the total draw time needed based on the current cost.
    	
    	In an animation, the system will expand as time progresses.
    	Each F command that draws a segment has a cost that depletes time.
    	To calculate the total amount of time for a number of generations,
    	we need to recurse through the system.
    	Time does not flow through the system linearly, 
    	it "branches" from generation to generation.
    	
    	"""
    	_ctx.push()
        self._reset()
        self._grow(generation, self.root, self.angle, self.d, draw=False)
        _ctx.pop()
        return max(self._duration, 0.1)

def lsystem(angle=20, segmentlength=40, rules={}, root=None):
	
	s = LSystem()
	s.angle = angle
	s.segmentlength = segmentlength
	if len(rules) > 0: 
		s.rules = rules
	if root: 
		s.root = root
	return s
	
create = lsystem

### TREE RULESETS ####################################################################################

def oak(angle=20, segmentlength=40):
	rules = {
  		"1" : "FF[++FF[2][+FF2]][-FF3]",
		"2" : "F-F-F+[2]F+F+F+F+([2]",
		"3" : "F+F+F-[2]F-F-F-F-([2]"
	}
	return lsystem(angle, segmentlength, rules) 
	
def beech(angle=15, segmentlength=40):
	rules = {
		"1" : "FF[-2][3][+3]",
		"2" : "FF+F-F-F[FFF3][+3]-F-F3",
		"3" : "FF-F+F+F[2][-2]+F+F2"
	}
	return lsystem(angle, segmentlength, rules)  

def yew(angle=20, segmentlength=40):
    rules = {
        "1" : "F-F+F[++2][F+2][F-2][--2]",
		"2" : "F+FF-F[++3][+3][-4][--4]",
		"3" : "-[4]F-FF-FF-FF-F[4]",
		"4" : "+[3]F+FF+FF+FF+F[3]"
    }
    return lsystem(angle, segmentlength, rules) 

def elm(angle=20, segmentlength=40):
    rules = {
        "1" : "F-F+F+F[(+2]-FF-F[(-2](2",
		"2" : "F+F-FF[(+1]-FFF[(-1](1",
    }
    return lsystem(angle, segmentlength, rules) 

def hawthorn(angle=20, segmentlength=40):
	rules = {
        "1" : "F-F+2",
		"2" : "F-[[-F-F+F+FF2]+FF2]+F[+F+F+FF2]-FF+F-F2"        
    }
	return lsystem(angle, segmentlength, rules) 

def eucalyptus(angle=23, segmentlength=40):
	rules = {
		"1" : "FFF+[2]F+(>[---1]",
		"2" : "FFF[1]+[1]+[1]+[1]"
    }
	return lsystem(angle, segmentlength, rules) 

def acacia(angle=25, segmentlength=40):
	
	rules = {
    	"1" : "FFF-[-F+F[2]-[1]]+[+F+F[1]-[1]]",
		"2" : "FF-[-F+F]+[+F+F2]"        
    }
	return lsystem(angle, segmentlength, rules) 

gnarled = oak
tall    = beech
great   = yew 
old     = elm
crooked = hawthorn  
slender = eucalyptus
strong  = acacia

trees = [gnarled, tall, great, old, crooked, slender, strong]

######################################################################################################

#stroke(0)
#tree = strong()
#tree.draw(250, 500, 6)



#!/usr/bin/env python2
# -*- coding: iso-8859-1 -*-

from __future__ import print_function
try:
    import gi
except ImportError:
    import pgi
    pgi.install_as_gi()

gi.require_version('Gtk', '3.0')

import errno
import os
import sys

from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GtkSource
from gi.repository import Pango

import shoebot

# for gtksourceview/gtksourceview2 compatibility
# try:
#     import gtksourceview
#
# except ImportError:
#     from gi.repository import GtkSource
#
#     class gtksourceview_SourceBuffer(GtkSource.Buffer):
#         def set_highlight(self, bool):
#             return self.set_highlight_syntax(bool)
#
#     class gtksourceview_SourceLanguagesManager(GtkSource.LanguageManager):
#         def get_language_from_mime_type(self, mime_types):
#             lang_manager = GtkSource.LanguageManager.get_default()
#             lang_result = None
#             for lang_id in lang_manager.get_language_ids():
#                  lang = lang_manager.get_language(lang_id)
#                  if mime_types in lang.get_mime_types():
#                       lang_result = lang
#                       break
#
#             return lang_result
#
#     gtksourceview = gtksourceview2
#     gtksourceview.SourceBuffer = gtksourceview_SourceBuffer
#     gtksourceview.SourceView = GtkSource.View
#     gtksourceview.SourceLanguagesManager = gtksourceview_SourceLanguagesManager
#     del gtksourceview_SourceBuffer
#     del gtksourceview_SourceLanguagesManager


from shoebot.data import ShoebotError


APP = 'shoebot'
DIR = sys.prefix + '/share/shoebot/locale'
RESPONSE_FORWARD = 0
RESPONSE_BACKWARD = 1


import locale
import gettext
locale.setlocale(locale.LC_ALL, '')
gettext.bindtextdomain(APP, DIR)
#gettext.bindtextdomain(APP)
gettext.textdomain(APP)
_ = gettext.gettext


if sys.platform != 'win32':
    ICON_FILE = '/usr/share/shoebot/icon.png'
else:
    import os.path
    ICON_FILE = os.path.join(sys.prefix, 'share', 'shoebot', 'icon.png')

def hsv_to_rgb(h, s, v):
    if s == 0.0:
        return (v, v, v)
    else:
        hue = h * 6.0
        saturation = s
        value = v

        if hue >= 6.0:
            hue = 0.0

        f = hue - int(hue)
        p = value * (1.0 - saturation)
        q = value * (1.0 - saturation * f)
        t = value * (1.0 - saturation * (1.0 - f))

        ihue = int(hue)
        if ihue == 0:
            return(value, t, p)
        elif ihue == 1:
                return(q, value, p)
        elif ihue == 2:
                return(p, value, t)
        elif ihue == 3:
            return(p, q, value)
        elif ihue == 4:
            return(t, p, value)
        elif ihue == 5:
            return(value, p, q)

def hue_to_color(hue):
    if hue > 1.0:
        raise ValueError

    h, s, v = hsv_to_rgb (hue, 1.0, 1.0)
    return (h*65535, s*65535, v*65535)


class Buffer(GtkSource.Buffer):
    N_COLORS = 16
    PANGO_SCALE = 1024

    def __init__(self):
        GObject.GObject.__init__(self)
        tt = self.get_tag_table()
        self.refcount = 0
        self.filename = None
        self.untitled_serial = -1
        self.color_tags = []
        self.color_cycle_timeout_id = 0
        self.start_hue = 0.0

        for i in range(Buffer.N_COLORS):
            tag = self.create_tag()
            self.color_tags.append(tag)

        #self.invisible_tag = self.create_tag(None, invisible=True)
        self.not_editable_tag = self.create_tag(editable=False,
                                                foreground="purple")
        self.found_text_tag = self.create_tag(foreground="red")

        tabs = Pango.TabArray.new(4, True)
        tabs.set_tab(0, Pango.TabAlign.LEFT, 10)
        tabs.set_tab(1, Pango.TabAlign.LEFT, 30)
        tabs.set_tab(2, Pango.TabAlign.LEFT, 60)
        tabs.set_tab(3, Pango.TabAlign.LEFT, 120)
        self.custom_tabs_tag = self.create_tag(tabs=tabs, foreground="green")
        TestText.buffers.push(self)

    def pretty_name(self):
        if self.filename:
            return os.path.basename(self.filename)
        else:
            if self.untitled_serial == -1:
                self.untitled_serial = TestText.untitled_serial
                TestText.untitled_serial += 1

            if self.untitled_serial == 1:
                return _('Untitled')
            else:
                return _('Untitled #%d') % self.untitled_serial

    def filename_set(self):
        for view in TestText.views:
            if view.text_view.get_buffer() == self:
                view.set_view_title()

    def search(self, str, view, forward):
        # remove tag from whole buffer
        start, end = self.get_bounds()
        self.remove_tag(self.found_text_tag, start, end)

        iter = self.get_iter_at_mark(self.get_insert())

        i = 0
        if str:
            if forward:
                while 1:
                    res = iter.forward_search(str, Gtk.TextSearchFlags.TEXT_ONLY)
                    if not res:
                        break
                    match_start, match_end = res
                    i += 1
                    self.apply_tag(self.found_text_tag, match_start, match_end)
                    iter = match_end
            else:
                while 1:
                    res = iter.backward_search(str, Gtk.TextSearchFlags.TEXT_ONLY)
                    if not res:
                        break
                    match_start, match_end = res
                    i += 1
                    self.apply_tag(self.found_text_tag, match_start, match_end)
                    iter = match_start

        dialog = Gtk.MessageDialog(view,
                                   Gtk.DialogFlags.DESTROY_WITH_PARENT,
                                   Gtk.MessageType.INFO,
                                   Gtk.ButtonsType.OK,
                                   _('%d strings found and marked in red') % i)

        dialog.connect("response", lambda x,y: dialog.destroy())

        dialog.show()

    def search_forward(self, str, view):
        self.search(str, view, True)

    def search_backward(self, str, view):
        self.search(str, view, False)

    def ref(self):
        self.refcount += 1

    def unref(self):
        self.refcount -= 1
        if self.refcount == 0:
            self.set_colors(False)
            TestText.buffers.remove(self)
            del self

    def color_cycle_timeout(self):
        self.cycle_colors()
        return True

    def set_colors(self, enabled):
        hue = 0.0

        if (enabled and self.color_cycle_timeout_id == 0):
            self.color_cycle_timeout_id = Gtk.timeout_add(
                200, self.color_cycle_timeout)
        elif (not enabled and self.color_cycle_timeout_id != 0):
            Gtk.timeout_remove(self.color_cycle_timeout_id)
            self.color_cycle_timeout_id = 0

        for tag in self.color_tags:
            if enabled:
                color = apply(TestText.colormap.alloc_color,
                              hue_to_color(hue))
                tag.set_property("foreground_gdk", color)
            else:
                tag.set_property("foreground_set", False)
            hue += 1.0 / Buffer.N_COLORS

    def cycle_colors(self):
        hue = self.start_hue

        for tag in self.color_tags:
            color = apply(TestText.colormap.alloc_color,
                          hue_to_color (hue))
            tag.set_property("foreground_gdk", color)

            hue += 1.0 / Buffer.N_COLORS
            if hue > 1.0:
                hue = 0.0

        self.start_hue += 1.0 / Buffer.N_COLORS
        if self.start_hue > 1.0:
            self.start_hue = 0.0

    def tag_event_handler(self, tag, widget, event, iter):
        char_index = iter.get_offset()
        tag_name = tag.get_property("name")
        if event.type == Gdk.MOTION_NOTIFY:
            print("Motion event at char %d tag `%s'\n" % (char_index, tag_name))
        elif event.type == Gdk.EventType.BUTTON_PRESS:
            print("Button press at char %d tag `%s'\n" % (char_index, tag_name))
        elif event.type == Gdk._2BUTTON_PRESS:
            print("Double click at char %d tag `%s'\n" % (char_index, tag_name))
        elif event.type == Gdk._3BUTTON_PRESS:
            print("Triple click at char %d tag `%s'\n" % (char_index, tag_name))
        elif event.type == Gdk.BUTTON_RELEASE:
            print("Button release at char %d tag `%s'\n" % (char_index, tag_name))
        elif (event.type == Gdk.KEY_PRESS or
              event.type == Gdk.KEY_RELEASE):
            print("Key event at char %d tag `%s'\n" % (char_index, tag_name))
        return False

    def fill_file_buffer(self, filename):
        try:
            f = open(filename, "r")
        except IOError, (errnum, errmsg):
            err = "Cannot open file '%s': %s" % (filename, errmsg)
            view = TestText.active_window_stack.get()
            dialog = Gtk.MessageDialog(view, Gtk.DialogFlags.MODAL,
                                       Gtk.MessageType.INFO,
                                       Gtk.ButtonsType.OK, err);
            result = dialog.run()
            dialog.destroy()
            return False

        iter = self.get_iter_at_offset(0)
        buf = f.read()
        f.close()
        self.set_text(buf)
        self.set_modified(False)

        return True

    def save_buffer(self):
        result = False
        have_backup = False
        if not self.filename:
            return False

        bak_filename = self.filename + "~"
        try:
            if os.path.isfile(bak_filename):
                os.remove(bak_filename)
            os.rename(self.filename, bak_filename)
        except (OSError, IOError), (errnum, errmsg):
            if errnum != errno.ENOENT:
                err = "Cannot back up '%s' to '%s': %s" % (self.filename,
                                                           bak_filename,
                                                           errmsg)
                view = TestText.active_window_stack.get()
                dialog = Gtk.MessageDialog(view, Gtk.DialogFlags.MODAL,
                                           Gtk.MessageType.INFO,
                                           Gtk.ButtonsType.OK, err);
                dialog.run()
                dialog.destroy()
                return False

        have_backup = True
        start, end = self.get_bounds()
        chars = self.get_slice(start, end, False)
        try:
            file = open(self.filename, "w")
            file.write(chars)
            file.close()
            result = True
            self.set_modified(False)
        except IOError, (errnum, errmsg):
            err = "Error writing to '%s': %s" % (self.filename, errmsg)
            view = TestText.active_window_stack.get()
            dialog = Gtk.MessageDialog(view, Gtk.DialogFlags.MODAL,
                                       Gtk.MessageType.INFO,
                                       Gtk.ButtonsType.OK, err);
            dialog.run()
            dialog.destroy()

        if not result and have_backup:
            try:
                os.rename(bak_filename, self.filename)
            except OSError, (errnum, errmsg):
                err = "Can't restore backup file '%s' to '%s': %s\nBackup left as '%s'" % (
                    self.filename, bak_filename, errmsg, bak_filename)
                view = TestText.active_window_stack.get()
                dialog = Gtk.MessageDialog(view, Gtk.DialogFlags.MODAL,
                                           Gtk.MessageType.INFO,
                                           Gtk.ButtonsType.OK, err);
                dialog.run()
                dialog.destroy()

        return result

    def confirm_overwrite_callback(self, chooser):
      uri = chooser.get_uri()
      if os.path.exists(self.filename):
          if os.path.exists(self.filename):
              if user_wants_to_replace_read_only_file (uri):
                  return Gtk.FILE_CHOOSER_CONFIRMATION_ACCEPT_FILENAME
              else:
                  return Gtk.FILE_CHOOSER_CONFIRMATION_SELECT_AGAIN
          else:
              # fall back to the default dialog
              return Gtk.FILE_CHOOSER_CONFIRMATION_CONFIRM

    def save_as_buffer(self):
        """
        Return True if the buffer was saved
        """
        chooser = ShoebotFileChooserDialog(_('Save File'), None, Gtk.FileChooserAction.SAVE,
            (Gtk.STOCK_SAVE, Gtk.ResponseType.ACCEPT,
             Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL))
        chooser.set_do_overwrite_confirmation(True)

        chooser.connect("confirm-overwrite", self.confirm_overwrite_callback)
        
        saved = chooser.run() == Gtk.ResponseType.ACCEPT
        if saved:
                old_filename = self.filename
                self.filename = chooser.get_filename()
                if self.save_buffer():
                    self.filename = chooser.get_filename()
                    self.filename_set()
                else:
                    self.filename = old_filename
        chooser.destroy()
        return saved

    def check_buffer_saved(self):
        """
        If the buffer was not saved then give the user the chance to save it
        or cancel.
        
        Return True is the buffer was saved in the end
        """
        if self.get_modified():
            pretty_name = self.pretty_name()
            msg = _("Save changes to '%s'?") % pretty_name
            view = TestText.active_window_stack.get()
            dialog = Gtk.MessageDialog(view, Gtk.DialogFlags.MODAL,
                                       Gtk.MessageType.QUESTION,
                                       Gtk.ButtonsType.YES_NO, msg);
            dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
            result = dialog.run()
            dialog.destroy()
            if result == Gtk.ResponseType.YES:
                if self.filename:
                    return self.save_buffer()
                else:
                    return self.save_as_buffer()
            elif result == Gtk.ResponseType.NO:
                return True
            else:
                return False
        else:
            return True


class ShoebotFileChooserDialog (Gtk.FileChooserDialog):

    CWD = None

    def __init__(self, *args, **kwargs):
        super(ShoebotFileChooserDialog, self).__init__ (*args, **kwargs)

        # set some defaults
        self.set_default_response (Gtk.ResponseType.OK)
        self.set_property('do-overwrite-confirmation', True)

        # set the working directory if available
        if ShoebotFileChooserDialog.CWD is not None:
            self.set_current_folder(ShoebotFileChooserDialog.CWD)

    def run(self):
        response = super (ShoebotFileChooserDialog, self).run()

        # get the working directory if the user clicked accepted the action
        if response == Gtk.ResponseType.ACCEPT:
            ShoebotFileChooserDialog.CWD = self.get_current_folder()

        return response


class ConsoleWindow:
    def __init__(self):
        # we define a scrollable window with automatic behavior for scrolling bars             
        self.text_window = Gtk.ScrolledWindow()
        self.text_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.text_area = Gtk.TextView()
        # set text area as editable in order to let it be cleaned by selecting and deleting
        # then we set wrap mode for text
        self.text_area.set_editable(True)
        self.text_area.set_wrap_mode(Gtk.WrapMode.WORD)
        self.text_buffer = self.text_area.get_buffer()        
        self.text_window.add(self.text_area)
        # here we set default values for background and text of console window
        self.text_area.modify_base(Gtk.StateType.NORMAL, Gdk.color_parse("dark grey"))
        self.text_area.modify_text(Gtk.StateType.NORMAL, Gdk.color_parse("red"))
        # then we define some text tag for defining colors for system messages and stdout
        self.tag_table = self.text_buffer.get_tag_table()
        ##self.stdout_tag = Gtk.TextTag("stdout")
        ##self.stdout_tag.set_property("foreground", "black")
        ##self.stdout_tag.set_property("style", "normal")
        #self.stdout_tag.set_property("weight", 600)   
        #self.stdout_tag.set_property("size-points", 9)

        self.stdout_tag = self.text_buffer.create_tag("system", foreground="black", weight=600, size_points=9)
        self.system_message_tag = self.text_buffer.create_tag("system", foreground="white")

        ##self.tag_table.add(self.stdout_tag)
        ##self.system_message_tag = Gtk.TextTag("system")
        ##self.system_message_tag.set_property("foreground", "white")
        ##self.system_message_tag.set_property("style", "normal")
        ##self.tag_table.add(self.system_message_tag)
        #self.text_area.modify_font(Pango.FontDescription("monospace italic 9"))


    def write(self, data, output=None, system=None):
        self.message = data
        if not output:
            # no tags set for stderr messages, color will be the one set for TextView            
            self.text_buffer.insert_at_cursor(self.message)
            self.message = ""
        elif system:
            # if output and system values are set, text is treated as a system message
            # and system tag is used
            self.iter = self.text_buffer.get_iter_at_mark(self.text_buffer.get_insert())
            self.text_buffer.insert_with_tags_by_name(self.iter, self.message, "system")
            self.message = ""
        else:
            # if only output value is set, tag used will be stdout
            self.iter = self.text_buffer.get_iter_at_mark(self.text_buffer.get_insert())
            self.text_buffer.insert_with_tags_by_name(self.iter, self.message, "stdout")
            self.message = ""
        # this is the trick to make gtk refresh the window                    
        while Gtk.events_pending():
            Gtk.main_iteration()


class Stdout_Filter(object):
    def __init__(self, parent):
        self.parent = parent
    def write(self, data):
        self.message = data
        self.parent.write(self.message, True)
        self.message = None

    
class View(Gtk.Window):
    ## Gtk3 TODO - GObject.type_register(ShoebotFileChooserDialog)
    FONT = None
    
    def __init__(self, buffer=None):
        menu_items = [
            ( _("/_File"), None, None, 0, "<Branch>" ),
            ( _("/File/_New"), "<control>N", self.do_new, 0, None ),
            ( _("/File/_Open"), "<control>O", self.do_open, 0, None ),
            ( _("/File/_Save"), "<control>S", self.do_save, 0, None ),
            ( _("/File/Save _As..."), None, self.do_save_as, 0, None ),
            ( _("/File/sep1"), None, None, 0, "<Separator>" ),
            ( _("/File/_Close"), "<control>W" , self.do_close, 0, None ),
            ( _("/File/E_xit"), "<control>Q" , self.do_exit, 0, None ),
            ( _("/_Edit"), None, None, 0, "<Branch>" ),
            ( _("/Edit/Undo"), "<control>Z", self.do_undo, 0, None ),
            ( _("/Edit/Redo"), "<control><shift>Z", self.do_redo, 0, None ),
            ( _("/Edit/sep1"), None, None, 0, "<Separator>" ),
            ( _("/Edit/Find..."), "<control>F", self.do_search, 0, None ),
            ( _("/_Settings"), None, None, 0, "<Branch>" ),
            ( _("/Settings/Wrap _Off"), None, self.do_wrap_changed, Gtk.WrapMode.NONE, "<RadioItem>" ),
            ( _("/Settings/Wrap _Words"), None, self.do_wrap_changed, Gtk.WrapMode.WORD, _("/Settings/Wrap Off") ),
            ( _("/Settings/Wrap _Chars"), None, self.do_wrap_changed, Gtk.WrapMode.CHAR, _("/Settings/Wrap Off") ),
            ( _("/_Run"), None, None, 0, "<Branch>" ),
            ( _("/Run/_Run script"), "<control>R", self.run_script, False, None ),
            ( _("/Run/sep1"), None, None, 0, "<Separator>" ),
            ( _("/Run/Run socket server"), None, self.do_socketserver_changed, True, "<ToggleItem>"),
            ( _("/Run/Show variables window"), None, self.do_varwindow_changed, True, "<ToggleItem>"),
            ( _("/Run/Run fullscreen"), None, self.do_fullscreen_changed, True, "<ToggleItem>"),
            ( _("/_Help"), None, None, 0, "<Branch>" ),
            ( _("/Help/_About"), None, self.do_about, 0, None )  
            ]

        if not buffer:
            buffer = Buffer()
        GObject.GObject.__init__(self)

        TestText.views.push(self)

        buffer.ref()

        ## Gtk3.TODO
        ##if not TestText.colormap:
        ##    TestText.colormap = self.get_colormap()

        self.connect("delete_event", self.delete_event_cb)

        self.accel_group = Gtk.AccelGroup()
        ##self.item_factory = Gtk.ItemFactory(Gtk.MenuBar, "<main>",
        ##                                self.accel_group)
        ##self.item_factory.set_data("view", self)
        ##self.item_factory.create_items(menu_items)

        self.add_accel_group(self.accel_group)
        
        hpaned = Gtk.HPaned()
        vbox = Gtk.VBox(False, 0)
        hpaned.add1(vbox)
        self.add(hpaned)

        ##vbox.pack_start(self.item_factory.get_widget("<main>", True, True, 0),
        ##                False, False, 0)

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        ## self.text_view = gtksourceview.SourceView(buffer)
        self.text_view = GtkSource.View.new_with_buffer(buffer)
        self.text_view.set_wrap_mode(Gtk.WrapMode.WORD)
        self.text_view.set_show_line_numbers(True)
        self.text_view.set_auto_indent(True)
        self.text_view.set_insert_spaces_instead_of_tabs(True)
        # FIXME: the following 2 lines don't have any effect :/
        self.text_view.tab_width = 4
        self.text_view.indent_width = 4
        ##self.text_view.connect("expose_event", self.tab_stops_expose)

        self.bhid = buffer.connect("mark_set", self.cursor_set_callback)
        
        if View.FONT is None:
            # Get font or fallback
            context = self.text_view.get_pango_context()
            fonts = context.list_families()
            for font in fonts:
                if font.get_name() == 'Bitstream Vera Sans Mono':
                    View.FONT = 'Bitstream Vera Sans Mono 8'
                    break
            else:
                print('Bitstream Vera Font not found.')
                print('Download and install it from here')
                print('http://ftp.gnome.org/pub/GNOME/sources/ttf-bitstream-vera/1.10/')
                View.FONT = 'Mono 8'

        ##self.text_view.modify_font(Pango.FontDescription(View.FONT))

        vbox.pack_start(sw, True, True, 0)
        sw.add(self.text_view)

        # this creates a console error and output window besides script window
        self.console_error = ConsoleWindow()
        # we create an instance for stdout filter
        self.stdout_filter = Stdout_Filter(self.console_error)
        # we redirect stderr
        ##sys.stderr = self.console_error
        # stdout is redirected too, but through the filter in order to get different color for text
        ##sys.stdout = self.stdout_filter
        # error-console window is added to container as second child
        hpaned.add2(self.console_error.text_window)
        hpaned.set_position(450)
        # message displayed in console-error window at start, the double true values passed makes it render with system message tag
        self.console_error.write(_("This is a console window, error messages and script output are shown here,\n you can clear the window selecting and deleting the content\n\n"), True, True)        
                
        self.set_default_size(800, 500)
        self.text_view.grab_focus()

        self.set_view_title()
        self.init_menus()

        # options toggle
        self.use_varwindow = False
        self.use_socketserver = False
        self.go_fullscreen = False

        # setup syntax highlighting
        manager = GtkSource.LanguageManager()
        ##language = manager.get_language_from_mime_type("text/x-python")
        language = manager.guess_language(None, "text/x-python")
        buffer.set_language(language)
        ##buffer.set_highlight(True)

        self.shoebot_window = None

        try:
            self.set_icon_from_file(ICON_FILE)
        except GObject.GError:
            # icon not found = no icon
            pass

        self.show_all()

    def delete_event_cb(self, window, event, data=None):
        TestText.active_window_stack.push(self)
        self.check_close_view()
        TestText.active_window_stack.pop()
        return True
    #
    # Menu callbacks
    #
    def get_empty_view(self):
        buffer = self.text_view.get_buffer()
        if (not buffer.filename and not buffer.get_modified()):
            return self
        else:
            return View(Buffer())

    def view_from_widget(widget):
        if isinstance(widget, Gtk.MenuItem):
            item_factory = Gtk.item_factory_from_widget(widget)
            return item_factory.get_data("view")
        else:
            app = widget.get_toplevel()
            return app.get_data("view")

    def do_new(self, callback_action, widget):
        View()

    def open_ok_func(self, filename):
        new_view = self.get_empty_view()
        buffer = new_view.text_view.get_buffer()
        if not buffer.fill_file_buffer(filename):
            if new_view != self:
                new_view.close_view()
            return False
        else:
            buffer.filename = filename
            buffer.filename_set()
            return True

    def do_open(self, callback_action, widget):
        chooser = ShoebotFileChooserDialog('Open File', None, Gtk.FileChooserAction.OPEN,
            (Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT,
             Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL))

        if chooser.run() == Gtk.ResponseType.ACCEPT:
            self.open_ok_func(chooser.get_filename())
        chooser.destroy()


    def do_save_as(self, callback_action, widget):
        TestText.active_window_stack.push(self)
        self.text_view.get_buffer().save_as_buffer()
        TestText.active_window_stack.pop()

    def do_save(self, callback_action, widget):
        TestText.active_window_stack.push(self)
        buffer = self.text_view.get_buffer()
        if not buffer.filename:
            self.do_save_as(None, callback_action)
        else:
            buffer.save_buffer()
            TestText.active_window_stack.pop()

    def do_close(self, callback_action, widget):
        TestText.active_window_stack.push(self)
        self.check_close_view()
        TestText.active_window_stack.pop()

    def do_exit(self, callback_action, widget):
        TestText.active_window_stack.push(self)
        for tmp in TestText.buffers:
            if not tmp.check_buffer_saved():
                return
        if hasattr(self, 'sbot_window'):
            self.sbot_window.finish()
            self.sbot_window.destroy()

        Gtk.main_quit()
        TestText.active_window_stack.pop()
        import sys
        sys.exit()

    def do_insert_and_scroll(self, callback_action, widget):
        buffer = self.text_view.get_buffer()

        start, end = buffer.get_bounds()
        mark = buffer.create_mark(None, end, False)

        buffer.insert(end,
                      "Hello this is multiple lines of text\n"
                      "Line 1\n"  "Line 2\n"
                      "Line 3\n"  "Line 4\n"
                      "Line 5\n")

        self.text_view.scroll_to_mark(mark, 0, True, 0.0, 1.0)
        buffer.delete_mark(mark)

    def do_wrap_changed(self, callback_action, widget):
        self.text_view.set_wrap_mode(callback_action)

    def do_varwindow_changed(self, callback_action, widget):
        self.use_varwindow = not self.use_varwindow

    def do_socketserver_changed(self, callback_action, widget):
        self.use_socketserver = not self.use_socketserver
        
    def do_fullscreen_changed(self, callback_action, widget):
        self.go_fullscreen = not self.go_fullscreen

    def do_color_cycle_changed(self, callback_action, widget):
        self.text_view.get_buffer().set_colors(callback_action)

    def do_apply_tabs(self, callback_action, widget):
        buffer = self.text_view.get_buffer()
        bounds = buffer.get_selection_bounds()
        if bounds:
            start, end = bounds
            if callback_action:
                buffer.remove_tag(buffer.custom_tabs_tag, start, end)
            else:
                buffer.apply_tag(buffer.custom_tabs_tag, start, end)

    def do_apply_colors(self, callback_action, widget):
        buffer = self.text_view.get_buffer()
        bounds = buffer.get_selection_bounds()
        if bounds:
            start, end = bounds
            if not callback_action:
                for tag in buffer.color_tags:
                    buffer.remove_tag(tag, start, end)
            else:
                tmp = buffer.color_tags
                i = 0
                next = start.copy()
                while next.compare(end) < 0:
                    next.forward_chars(2)
                    if next.compare(end) >= 0:
                        next = end

                    buffer.apply_tag(tmp[i], start, next)
                    i += 1
                    if i >= len(tmp):
                        i = 0
                    start = next.copy()

    def do_remove_tags(self, callback_action, widget):
        buffer = self.text_view.get_buffer()
        bounds = buffer.get_selection_bounds()
        if bounds:
            start, end = bounds
            buffer.remove_all_tags(start, end)

    def dialog_response_callback(self, dialog, response_id):
        if (response_id != RESPONSE_FORWARD and
            response_id != RESPONSE_BACKWARD):
            dialog.destroy()
            return

        start, end = dialog.buffer.get_bounds()
        search_string = start.get_text(end)

        print(_("Searching for `%s'\n") % search_string)

        buffer = self.text_view.get_buffer()
        if response_id == RESPONSE_FORWARD:
            buffer.search_forward(search_string, self)
        elif response_id == RESPONSE_BACKWARD:
            buffer.search_backward(search_string, self)

        dialog.destroy()

    def do_search(self, callback_action, widget):
        search_text = Gtk.TextView()
        dialog = Gtk.Dialog(_("Search"), self,
                            Gtk.DialogFlags.DESTROY_WITH_PARENT,
                            (_("Forward"), RESPONSE_FORWARD,
                             _("Backward"), RESPONSE_BACKWARD,
                             Gtk.STOCK_CANCEL, Gtk.ResponseType.NONE))
        dialog.vbox.pack_end(search_text, True, True, 0)
        dialog.buffer = search_text.get_buffer()
        dialog.connect("response", self.dialog_response_callback)

        search_text.show()
        search_text.grab_focus()
        dialog.show_all()

    def do_undo(self, callback_action, widget):
        buffer = self.text_view.get_buffer()
        if buffer.can_undo():
            buffer.undo()

    def do_redo(self, callback_action, widget):
        buffer = self.text_view.get_buffer()
        if buffer.can_redo():
            buffer.redo()
            
    #def do_about(self, callback_action, widget):
        #about = '''Shoebot is a pure Python graphics robot: 
#it takes a Python script as input, which describes a drawing process,
#and outputs a graphic in a common open standard format (SVG, PDF, PostScript, or PNG).\n
#It has a simple text editor GUI, and scripts can describe their own GUIs for
#controlling variables interactively.\n\n
        #Thanks to:\n
        #Ricardo Lafuente <r AT sollec.org>
        #Francesco Fantoni<francesco AT hv-a.com>
        #Dave Crossland <dave AT lab6.com>
        #Stuart Axon <stuaxo AT gmail.com>
        #Paulo Silva <nitrofurano AT gmail.com>
        #Tetsuya Saito <t2psyto AT gmail.com>\n
        #http://shoebot.net/\n
        #Version: 0.4-beta                
        #'''
        #self.console_error.write(about)            

    def on_url(self, d, link, data):
        import webbrowser
        webbrowser.open_new(data)
        
    def do_about(self, callback_action, widget):
        # about dialog
        dlg = Gtk.AboutDialog()
        self.website = "http://shoebot.net/"
        self.authors = ["Dave Crossland <dave AT lab6.com>", "est <electronixtar AT gmail.com>", "Francesco Fantoni <francesco AT hv-a.com>", "Paulo Silva <nitrofurano AT gmail.com>", "Pedro Angelo <pangelo AT virii-labs.org>", "Ricardo Lafuente <ricardo AT sollec.org>", "Stuart Axon <stuaxo2 AT yahoo.com>", "Tetsuya Saito <t2psyto AT gmail.com>"]
        Gtk.about_dialog_set_url_hook(self.on_url, self.website)
        dlg.set_version("0.4")
        dlg.set_name("shoebot")
        # TODO: add license text
        dlg.set_license("GPLv3")
        dlg.set_authors(self.authors)
        dlg.set_website(self.website)
        def close(w, res):
            if res == Gtk.ResponseType.CANCEL:
                    w.hide()
        dlg.connect("response", close)
        dlg.run()
            

    def init_menus(self):
        text_view = self.text_view
        direction = text_view.get_direction()
        wrap_mode = text_view.get_wrap_mode()
        menu_item = None

        ##if direction == Gtk.TextDirection.LTR:
        ##    menu_item = self.item_factory.get_widget("/Settings/Left-to-Right")
        ##elif direction == Gtk.TextDirection.RTL:
        ##    menu_item = self.item_factory.get_widget("/Settings/Right-to-Left")

        if menu_item:
            menu_item.activate()

        ##if wrap_mode == Gtk.WrapMode.NONE:
        ##    menu_item = self.item_factory.get_widget("/Settings/Wrap Off")
        ##elif wrap_mode == Gtk.WrapMode.WORD:
        ##    menu_item = self.item_factory.get_widget("/Settings/Wrap Words")
        ##elif wrap_mode == Gtk.WrapMode.CHAR:
        ##    menu_item = self.item_factory.get_widget("/Settings/Wrap Chars")

        ##if menu_item:
        ##    menu_item.activate()

    def close_view(self):
        TestText.views.remove(self)
        buffer = self.text_view.get_buffer()
        # buffer.unref()
        buffer.disconnect(self.bhid)
        self.text_view.destroy()
        del self.text_view
        self.text_view = None
        self.destroy()
        del self
        if not TestText.views:
            Gtk.main_quit()

    def check_close_view(self):
        buffer = self.text_view.get_buffer()
        if (buffer.refcount > 1 or
            buffer.check_buffer_saved()):
            self.close_view()

    def set_view_title(self):
        pretty_name = self.text_view.get_buffer().pretty_name()
        title = "Shoebot - " + pretty_name
        self.set_title(title)

    def cursor_set_callback(self, buffer, location, mark):

        # TODO: here should go the line syntax highlighter
        # 1. get buffer
        #      get modification state
        # 2. get line contents
        # 3. replace by pygmentised content
        #      revert to modification state
        pass

    def tab_stops_expose(self, widget, event):
        #print(self, widget, event)
        text_view = widget

        # See if this expose is on the tab stop window
        top_win = text_view.get_window(Gtk.TextWindowType.TOP)
        bottom_win = text_view.get_window(Gtk.TextWindowType.BOTTOM)

        if event.window == top_win:
            type = Gtk.TextWindowType.TOP
            target = top_win
        elif event.window == bottom_win:
            type = Gtk.TextWindowType.BOTTOM
            target = bottom_win
        else:
            return False

        first_x = event.area.x
        last_x = first_x + event.area.width

        first_x, y = text_view.window_to_buffer_coords(type, first_x, 0)
        last_x, y = text_view.window_to_buffer_coords(type, last_x, 0)

        buffer = text_view.get_buffer()
        insert = buffer.get_iter_at_mark(buffer.get_insert())
        attrs = Gtk.TextAttributes()
        insert.get_attributes(attrs)

        tabslist = []
        in_pixels = False
        if attrs.tabs:
            tabslist = attrs.tabs.get_tabs()
            in_pixels = attrs.tabs.get_positions_in_pixels()

        for align, position in tabslist:
            if not in_pixels:
                position = Pango.PIXELS(position)

            pos, y = text_view.buffer_to_window_coords(type, position, 0)
            target.draw_line(text_view.style.fg_gc[text_view.state],
                             pos, 0, pos, 15)

        return True

    def get_lines(self, first_y, last_y, buffer_coords, numbers):
        text_view = self.text_view
        # Get iter at first y
        iter, top = text_view.get_line_at_y(first_y)

        # For each iter, get its location and add it to the arrays.
        # Stop when we pass last_y
        count = 0
        size = 0

        while not iter.is_end():
            y, height = text_view.get_line_yrange(iter)
            buffer_coords.append(y)
            line_num = iter.get_line()
            numbers.append(line_num)
            count += 1
            if (y + height) >= last_y:
                break
            iter.forward_line()

        return count

    def run_script(self, callback_action, widget):
        # get the buffer contents
        buffer = self.text_view.get_buffer()
        start, end = buffer.get_bounds()
        codestring = buffer.get_text(start, end)
        try:
            if buffer.filename:
                os.chdir(os.path.dirname(buffer.filename))
                
                
            bot = shoebot.create_bot(codestring, 'NodeBox', server=self.use_socketserver, show_vars=self.use_varwindow, window = True)
            self.sbot_window = bot._canvas.sink
            bot.run(codestring, run_forever = True, iterations = None)
        except ShoebotError, NameError:
            import traceback
            import sys

            errmsg = traceback.format_exc(limit=1)
            err = "Error in Shoebot script:\n %s" % (errmsg)
            dialog = Gtk.MessageDialog(self, Gtk.DialogFlags.MODAL,
                                       Gtk.MessageType.INFO,
                                       Gtk.ButtonsType.OK, err);
            result = dialog.run()
            dialog.destroy()
            return False

        # TODO: have a try/except that shows an error window         

class Stack(list):
    def __init__(self):
        list.__init__(self)

    def push(self, item):
        self.insert(-1, item)

    def pop(self):
        del self[0]

    def get(self):
        return self[0]


class TestText(object):
    untitled_serial = 1
    colormap = None
    active_window_stack = Stack()
    buffers = Stack()
    views = Stack()

    def __init__(self, filelist):
        view = View()
        self.active_window_stack.push(view)
        for fname in filelist:
            filename = os.path.abspath(fname)
            view.open_ok_func(filename)
        self.active_window_stack.pop()

    def main(self):
        Gtk.main()
        return 0


def main():
    testtext = TestText(sys.argv[1:])
    testtext.main()

if __name__ == "__main__":
    main()



"""
gobject-based socket server from
http://roscidus.com/desktop/node/413

Handles interaction when connecting over a socket, e.g. telnet.
The actual handling of commands is passed to te shell module.

$ sbot -ws examples/basic/vars_boolean.bot &
$ telnet 127.0.0.1 7777

# Now enter the commands below and see the bot state change
color_switch=False
color_switch=True
"""

import sys
import socket
import gettext, locale

try:
    import gi
except ImportError:
    import pgi
    pgi.install_as_gi()

from gi.repository import GObject
from shell import ShoebotCmd

APP = 'shoebot'
DIR = sys.prefix + '/share/shoebot/locale'
locale.setlocale(locale.LC_ALL, '')
gettext.bindtextdomain(APP, DIR)
#gettext.bindtextdomain(APP)
gettext.textdomain(APP)
_ = gettext.gettext


# For now this is a bit basic, but works enough.


BANNER = """
############`             [Welcome to Shoebot Telnet Console.]`
##```####```              `````````````````````````````````````
############`
############`
##```````````
############`
`````````````

Set variables with  var=value

Enter vars to view variables.
Enter bye or press CTRL-D to quit.

"""

INTRO = "[o_o] " + '"Shoebot Telnet Shell, enter "help" for help."'

def create_listening_socket(host, port, handler):
    """
    Create socket and set listening options
    :param host:
    :param port:
    :param handler:
    :return:
    """
    sock = socket.socket()
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    sock.bind((host, port))
    sock.listen(1)

    GObject.io_add_watch(sock, GObject.IO_IN, handler)
    return sock

class SocketServer(object):
    def __init__(self, bot, host, port):
        '''Initialize server and start listening.'''
        create_listening_socket(host, port, self.listener)
        self.shell = None
        self.bot = bot

    def listener(self, sock, *args):
        '''Asynchronous connection listener. Starts a handler for each connection.'''
        conn, addr = sock.accept()
        f = conn.makefile(conn)
        self.shell = ShoebotCmd(self.bot, stdin=f, stdout=f, intro=INTRO)

        print(_("Connected"))
        GObject.io_add_watch(conn, GObject.IO_IN, self.handler)
        if self.shell.intro:
            self.shell.stdout.write(str(self.shell.intro)+"\n")
            self.shell.stdout.flush()
        return True

    def handler(self, conn, *args):
        '''
        Asynchronous connection handler. Processes each line from the socket.
        '''
        # lines from cmd.Cmd
        self.shell.stdout.write(self.shell.prompt)
        line = self.shell.stdin.readline()
        if not len(line):
            line = 'EOF'
            return False
        else:
            line = line.rstrip('\r\n')
            line = self.shell.precmd(line)
            stop = self.shell.onecmd(line)
            stop = self.shell.postcmd(stop, line)
            self.shell.stdout.flush()
            self.shell.postloop()
            # end lines from cmd.Cmd
            if stop:
                self.shell = None
                conn.close()
            return not stop



#!/usr/bin/env python2

# Shoebot setup script
#
# python setup.py install', or
#    python setup.py --help' for more options
#

from __future__ import print_function

import glob
import os
import platform
import shutil
import sys
import textwrap

is_pypy = '__pypy__' in sys.builtin_module_names
is_jython = platform.system == 'Java'

here = os.path.dirname(os.path.abspath(__file__))

try:
    from setuptools import setup
    from setuptools import Command as CleanBaseCommand
except ImportError:
    from distutils.core import setup
    from distutils.command import clean as CleanBaseCommand


class CleanCommand(CleanBaseCommand):
    """Custom clean command to tidy up the project root."""
    CLEAN_FILES = './build ./dist ./*.pyc ./*.tgz ./*.egg-info'.split(' ')

    user_options = []

    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        global here

        for path_spec in self.CLEAN_FILES:
            # Make paths absolute and relative to this path
            abs_paths = glob.glob(os.path.normpath(os.path.join(here, path_spec)))
            for path in [str(p) for p in abs_paths]:
                if not path.startswith(here):
                    # Die if path in CLEAN_FILES is absolute + outside this directory
                    raise ValueError("%s is not a path inside %s" % (path, here))
                print('removing %s' % os.path.relpath(path))
                shutil.rmtree(path)

long_description = textwrap.dedent("""
    Shoebot is a pure Python graphics robot: It takes a Python script as input,
    which describes a drawing process, and outputs a graphic in a common open
    standard format (SVG, PDF, PostScript, or PNG). It has a simple text editor
    GUI, and scripts can describe their own GUIs for controlling variables
    interactively. Being pure Python, it can also be used as a Python module,
    a plugin for Python-scriptable tools such as Inkscape, and run from the
    command line.
""")

# the following libraries will not be installed
EXCLUDE_LIBS = ['lib/sbopencv', 'lib/sbopencv/blobs']

for lib in EXCLUDE_LIBS:
    # get subdirs of excluded libs
    for root, dir, files in list(os.walk(lib))[1:]:
        EXCLUDE_LIBS.append(root)

# dir globbing approach taken from Mercurial's setup.py
datafiles = [(os.path.join('share/shoebot/', root), [os.path.join(root, file_)
                                                     for file_ in files]) for root, dir, files in os.walk('examples')]
datafiles.append(('share/pixmaps', ['assets/shoebot-ide.png']))
datafiles.append(('share/shoebot/data', ['assets/kant.xml']))
datafiles.append(('share/applications', ['assets/shoebot-ide.desktop']))

datafiles.extend([(os.path.join('share/shoebot/', root), [os.path.join(root, file_)
                                                          for file_ in files]) for root, dir, files in
                  os.walk('locale')])

# include all libs EXCEPT the ones mentioned in EXCLUDE_LIBS

datafiles.extend([(os.path.join('share/shoebot/', root) ,[os.path.join(root, file_)
for file_ in files]) for root,dir,files in os.walk('lib') if root not in EXCLUDE_LIBS])

if not is_pypy:
    NUMPY="numpy>=1.9.1"
else:
    NUMPY=""


# Also requires one of 'vext.gi' or 'pgi' to run in GUI
BASE_REQUIREMENTS=[
    "setuptools>=15.0.1",  #

    "cairocffi>=0.7.2",
    "meta",              # as of meta version 0.4.1 the version is borked when installing from pypi
    NUMPY,
    "Pillow==2.8.1",
    "pubsub==0.1.1",
]


# requirements to run examples
EXAMPLE_REQUIREMENTS=[
  "fuzzywuzzy==0.5.0",   # sbaudio
  "planar",              # examples
  "PySoundCard==0.5.0",  # sbaudio
]

def requirements(with_pgi=None, with_examples=True, debug=True):
    """
    Build requirements based on flags

    :param with_pgi: Use 'pgi' instead of 'gi' - False on CPython, True elsewhere
    :param with_examples:
    :return:
    """
    reqs = list(BASE_REQUIREMENTS)
    if with_pgi is None:
        if is_pypy or is_jython:
            with_pgi = True
        else:
            with_pgi = False

    if debug:
        print('setup options: ')
        print('with_pgi:      ', 'yes' if with_pgi else 'no')
        print('with_examples: ', 'yes' if with_examples else 'no')
    if with_pgi:
        reqs.append("pgi")
        if debug:
            print("warning, as of April 2015 typography does not work with pgi")
    else:
        reqs.append("vext.gi>=0.5.11")
    if with_examples:
        reqs.extend(EXAMPLE_REQUIREMENTS)

    if debug:
        print('')
        print('')
        for req in reqs:
            print(req)
    return reqs


setup(name="shoebot",
      version="1.2.1",
      description="Vector graphics scripting application",
      long_description=long_description,
      author="Ricardo Lafuente",
      author_email="r@manufacturaindependente.org",
      license='GPL v3',
      url="http://shoebot.net",
      cmdclass={
          'clean': CleanCommand,
      },
      packages=[
          "shoebot",
          "shoebot.core",
          "shoebot.data",
          "shoebot.gui",
          "shoebot.io",
          "shoebot.grammar",
          "shoebot.grammar.nodebox-lib",
          "shoebot.grammar.nodebox-lib.nodebox",
          "shoebot.grammar.nodebox-lib.nodebox.graphics",
          "shoebot.grammar.nodebox-lib.nodebox.geo"
      ],
      data_files=datafiles,
      install_requires=requirements(debug="install" in sys.argv),
      entry_points={
          "console_scripts": [
             "sbot=shoebot.run:main",
             "shoebot-ide=shoebot.gui.ide:main"
          ],
          "gui_scripts": "shoebot-ide=shoebot.gui.ide:main"
      }
)



#!/usr/bin/env python2
import os

try:
    import gi
except ImportError:
    import pgi
    pgi.install_as_gi()

from gi.repository import Gtk
from shoebot.core.events import EVENT_VARIABLE_UPDATED, publish_event
from shoebot.core.var_listener import VarListener
from pkg_resources import resource_filename, Requirement
ICON_FILE = resource_filename(Requirement.parse("shoebot"), "share/pixmaps/shoebot-ide.png")


NUMBER = 1
TEXT = 2
BOOLEAN = 3
BUTTON = 4


def pretty_name(name):
    return (name or '').replace('_', ' ').capitalize()


class VarWindow(object):
    def __init__(self, parent, bot, title = None):
        self.parent = parent
        self.bot = bot

        self.var_listener = VarListener(self)

        self.window = Gtk.Window()
        self.window.set_destroy_with_parent(True)
        self.window.connect("destroy", self.do_quit)

        if os.path.isfile(ICON_FILE):
            self.window.set_icon_from_file( ICON_FILE )

        self.container = Gtk.VBox(homogeneous=True, spacing=20)

        # set up sliders
        self.widgets = {}
        self.vars = {}
        self.add_variables()

        self.window.add(self.container)
        self.window.set_size_request(400, 35*len(self.widgets.keys()))
        self.window.show_all()

        if title:
            self.window.set_title(title)
        ## Gtk.main()

    def add_variables(self):
        """
        Add all widgets to specified vbox
        :param container:
        :return:
        """
        for v in sorted(self.bot._vars.values()):
            self.add_variable(v)

    def add_variable(self, v):
        if v.type is NUMBER:
            self.widgets[v.name] = self.add_number(v)
        elif v.type is TEXT:
            self.widgets[v.name] = self.add_text(v)
        elif v.type is BOOLEAN:
            self.widgets[v.name] = self.add_boolean(v)
        elif v.type is BUTTON:
            self.widgets[v.name] = self.add_button(v)
        else:
            raise ValueError('Unknown variable type.')
        self.vars[v.name] = v

    def add_number(self, v):
        # create a slider for each var
        sliderbox = Gtk.HBox(homogeneous=False, spacing=0)
        label = Gtk.Label(pretty_name(v.name))
        sliderbox.pack_start(label, False, True, 20)

        if v.min != v.max:
            step = v.step
        else:
            step = 0

        if v.max - v.min > 2:
            adj = Gtk.Adjustment(value=v.value, lower=v.min, upper=v.max, step_incr=step, page_incr=2, page_size=1)
        else:
            adj = Gtk.Adjustment(value=v.value, lower=v.min, upper=v.max, step_incr=step)
        adj.connect("value_changed", self.widget_changed, v)
        hscale = Gtk.HScale(adjustment=adj)
        hscale.set_value_pos(Gtk.PositionType.RIGHT)
        hscale.set_value(v.value)
        sliderbox.pack_start(hscale, True, True, 0)
        self.container.pack_start(sliderbox, True, True, 0)

        return hscale

    def add_text(self, v):
        textcontainer = Gtk.HBox(homogeneous=False, spacing=0)
        label = Gtk.Label(pretty_name(v.name))
        textcontainer.pack_start(label, False, True, 20)

        entry = Gtk.Entry()
        entry.set_text(v.value)
        entry.connect("changed", self.widget_changed, v)
        textcontainer.pack_start(entry, True, True, 0)
        self.container.pack_start(textcontainer, True, True, 0)

        return entry

    def add_boolean(self, v):
        buttoncontainer = Gtk.HBox(homogeneous=False, spacing=0)
        button = Gtk.CheckButton(label=pretty_name(v.name))
        button.set_active(v.value)
        # we send the state of the button to the callback method
        button.connect("toggled", self.widget_changed, v)

        buttoncontainer.pack_start(button, True, True, 0)
        self.container.pack_start(buttoncontainer, True, True, 0)

        return button

    def add_button(self, v):
        buttoncontainer = Gtk.HBox(homogeneous=False, spacing=0)
        # in buttons, the varname is the function, so we use __name__

        func_name = v.name

        def call_func(*args):
            func = self.bot._namespace[func_name]
            func()

        button = Gtk.Button(label=pretty_name(v.name))
        button.connect("clicked", call_func, None)
        buttoncontainer.pack_start(button, True, True, 0)
        self.container.pack_start(buttoncontainer, True, True, 0)

        return button

    def do_destroy(self, widget):
        self.var_listener.remove()

    def do_quit(self, widget):
        pass

    def update_var(self, name, value):
        """
        :return: success, err_msg_if_failed
        """
        widget = self.widgets.get(name)
        if widget is None:
            return False, 'No widget found matching, {}'.format(name)

        try:
            if isinstance(widget, Gtk.CheckButton):
                widget.set_active(value)
                return True, widget.get_active()
            elif isinstance(widget, Gtk.Entry):
                widget.set_text(value)
                return True, widget.get_text()
            else:
                widget.set_value(value)
                return True, widget.get_value()
        except Exception as e:
            return False, str(e)

    def widget_changed(self, widget, v):
        ''' Called when a slider is adjusted. '''
        # set the appropriate bot var
        if v.type is NUMBER:
            self.bot._namespace[v.name] = widget.get_value()
            self.bot._vars[v.name].value = widget.get_value()  ## Not sure if this is how to do this - stu
            publish_event(EVENT_VARIABLE_UPDATED, v) # pretty dumb for now
        elif v.type is BOOLEAN:
            self.bot._namespace[v.name] = widget.get_active()
            self.bot._vars[v.name].value = widget.get_active()  ## Not sure if this is how to do this - stu
            publish_event(EVENT_VARIABLE_UPDATED, v) # pretty dumb for now
        elif v.type is TEXT:
            self.bot._namespace[v.name] = widget.get_text()
            self.bot._vars[v.name].value = widget.get_text()  ## Not sure if this is how to do this - stu
            publish_event(EVENT_VARIABLE_UPDATED, v) # pretty dumb for now

    def var_added(self, v):
        """
        var was added in the bot while it ran, possibly
        by livecoding

        :param v:
        :return:
        """
        self.add_variable(v)

        self.window.set_size_request(400, 35*len(self.widgets.keys()))
        self.window.show_all()

    def var_deleted(self, v):
        """
        var was added in the bot

        :param v:
        :return:
        """
        widget = self.widgets[v.name]

        # widgets are all in a single container ..
        parent = widget.get_parent()
        self.container.remove(parent)
        del self.widgets[v.name]

        self.window.set_size_request(400, 35*len(self.widgets.keys()))
        self.window.show_all()

    def var_updated(self, v):
        pass





# Copyright (c) 2007 Tom De Smedt.
# See LICENSE.txt for details.

try: from en import wordnet
except:
    try: import wordnet
    except:
        pass

class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

#### GRAPH HOVER/CLICK/DRAG EVENTS ###################################################################

class events:
    
    def __init__(self, graph, _ctx):
        
        self.graph = graph
        self._ctx = _ctx
        
        # Can contain a node:
        self.hovered = None
        self.pressed = None
        self.dragged = None
        self.clicked = None
        
        # Displays when hovering over a node.
        self.popup = False
        self.popup_text = {}
    
    def copy(self, graph):
    
        """ Returns a copy of the event handler, remembering the last node clicked.
        """
    
        e = events(graph, self._ctx)
        e.clicked = self.clicked
        return e
    
    def _mouse(self):
        
        return Point(
            self._ctx._ns["MOUSEX"], 
            self._ctx._ns["MOUSEY"]
        )
        
    mouse = property(_mouse)
 
    def _mousedown(self):
        
        if self._ctx._ns["mousedown"]:
            return True
        else:
            return False

    mousedown = property(_mousedown)

    def update(self):
    
        """ Interacts with the graph by clicking or dragging nodes.
        Hovering a node fires the callback function events.hover().
        Clicking a node fires the callback function events.click().
        """
    
        if self.mousedown:
        
            # When not pressing or dragging, check each node.
            if not self.pressed and not self.dragged:
                for n in self.graph.nodes:
                    if self.mouse in n:
                        self.pressed = n
                        break
                    
            # If a node is pressed, check if a drag is started.
            elif self.pressed and not self.mouse in self.pressed:
                self.dragged = self.pressed
                self.pressed = None
            
            # Drag the node (right now only for springgraphs).
            elif self.dragged and self.graph.layout.type == "spring":
                self.drag(self.dragged)
                self.graph.layout.i = min(100, max(2, self.graph.layout.n-100))
    
        # Mouse is clicked on a node, fire callback.
        elif self.pressed and self.mouse in self.pressed:
            self.clicked = self.pressed
            self.pressed = None
            self.graph.layout.i = 2
            self.click(self.clicked)
    
        # Mouse up.
        else:
            self.hovered = None
            self.pressed = None
            self.dragged = None
        
            # Hovering over a node?
            for n in self.graph.nodes:
                if self.mouse in n:
                    self.hovered = n
                    self.hover(n)
                    break
    
    def drag(self, node):

        """ Drags given node to mouse location.
        """
    
        dx = self.mouse.x - self.graph.x
        dy = self.mouse.y - self.graph.y

        # A dashed line indicates the drag vector.
        s = self.graph.styles.default
        self._ctx.nofill()
        self._ctx.nostroke()
        if s.stroke: 
            self._ctx.strokewidth(s.strokewidth)
            self._ctx.stroke(
                s.stroke.r, 
                s.stroke.g, 
                s.stroke.g, 
                0.75
            )
        p = self._ctx.line(node.x, node.y, dx, dy, draw=False)
        try: p._nsBezierPath.setLineDash_count_phase_([2,4], 2, 50)
        except:
            pass
        self._ctx.drawpath(p)
        r = node.__class__(None).r * 0.75
        self._ctx.oval(dx-r/2, dy-r/2, r, r)
    
        node.vx = dx / self.graph.d
        node.vy = dy / self.graph.d
        
    def hover(self, node):
        
        """ Displays a popup when hovering over a node.
        """
        
        if self.popup == False: return
        if self.popup == True or self.popup.node != node:
            if self.popup_text.has_key(node.id):
                texts = self.popup_text[node.id]
            else:
                texts = None
            self.popup = popup(self._ctx, node, texts)
        self.popup.draw()

    def click(self, node):
        
        pass

### POPUP ############################################################################################   

class popup:
    
    """ An information box used when hovering over a node.
    It takes a list of alternating texts to display.
    """
    
    def __init__(self, _ctx, node, texts=None, width=200, speed=2.0):
        
        if texts != None and not isinstance(texts, (tuple, list)):
            texts = [texts]
        
        self._ctx = _ctx
        self.node = node
        
        self.i = 0 
        self.q = texts

        # When no texts were supplied, fall back to WordNet.
        # If WordNet is loaded, gather gloss descriptions for the node's id.
        if self.q == None:
            self.q = []
            try:
                id = str(self.node.id)
                for i in range(wordnet.count_senses(id)):
                    txt  = id + " | " + wordnet.gloss(id, sense=i)
                    self.q.append(txt)
            except:
                pass
    
        # Defaults for colors and typography.
        self.background = self._ctx.color(0.00, 0.10, 0.15, 0.60)
        self.text       = self._ctx.color(1.00, 1.00, 1.00, 0.80)
        self.font       = "Verdana"
        self.fontsize   = 9.5

        # Cached outlined versions of the text.
        self._textpaths = []
        self._w = width
        
        self.speed = speed
        self.delay = 20 / self.speed

        self.fi = 0  # current frame
        self.fn = 0  # frame count
        self.mf = 50 # minimum frame count

    def textpath(self, i):
        
        """ Returns a cached textpath of the given text in queue.
        """
        
        if len(self._textpaths) == i:
            self._ctx.font(self.font, self.fontsize)
            txt = self.q[i]
            if len(self.q) > 1:
                # Indicate current text (e.g. 5/13).
                txt += " ("+str(i+1)+"/" + str(len(self.q))+")"
            p = self._ctx.textpath(txt, 0, 0, width=self._w)
            h = self._ctx.textheight(txt, width=self._w)
            self._textpaths.append((p, h))

        return self._textpaths[i]
        
    def update(self):
        
        """ Rotates the queued texts and determines display time.
        """
        
        if self.delay > 0:
            # It takes a while for the popup to appear.
            self.delay -= 1; return
            
        if self.fi == 0:
            # Only one text in queue, displayed infinitely.
            if len(self.q) == 1: 
                self.fn = float("inf")
            # Else, display time depends on text length.
            else:
                self.fn = len(self.q[self.i]) / self.speed
                self.fn = max(self.fn, self.mf)            
            
        self.fi += 1
        if self.fi > self.fn:
            # Rotate to the next text in queue.
            self.fi = 0
            self.i = (self.i+1) % len(self.q)
        
    def draw(self):
        
        """ Draws a popup rectangle with a rotating text queue.        
        """ 
        
        if len(self.q) > 0:
            self.update()
            
            if self.delay == 0:
                
                # Rounded rectangle in the given background color.
                p, h = self.textpath(self.i)
                f = self.fontsize
                self._ctx.fill(self.background)
                self._ctx.rect(
                    self.node.x + f*1.0, 
                    self.node.y + f*0.5, 
                    self._w + f, 
                    h + f*1.5, 
                    roundness=0.2
                )
                
                # Fade in/out the current text.
                alpha = 1.0
                if self.fi < 5: 
                    alpha = 0.2 * self.fi
                if self.fn-self.fi < 5: 
                    alpha = 0.2 * (self.fn-self.fi)
                self._ctx.fill(
                    self.text.r,
                    self.text.g,
                    self.text.b,
                    self.text.a * alpha
                )
                
                self._ctx.translate(self.node.x + f*2.0, self.node.y + f*2.5)
                self._ctx.drawpath(p)